# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# upv.py
# Description: Implementation of UPV algorithm
# ============================================

import torch
from math import sqrt
from functools import partial
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from .network_model import UPVGenerator, UPVDetector
from utils.transformers_config import TransformersConfig
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization
from exceptions.exceptions import AlgorithmNameMismatchError, InvalidDetectModeError


class UPVConfig(BaseConfig):
    """Config class for UPV algorithm, load config file and initialize parameters."""

    # def __init__(self, algorithm_config: str, transformers_config: TransformersConfig) -> None:
    #     """
    #         Initialize the UPV configuration.

    #         Parameters:
    #             algorithm_config (str): Path to the algorithm configuration file.
    #             transformers_config (TransformersConfig): Configuration for the transformers model.
    #     """
    #     if algorithm_config is None:
    #         config_dict = load_config_file('config/UPV.json')
    #     else:
    #         config_dict = load_config_file(algorithm_config)
    #     if config_dict['algorithm_name'] != 'UPV':
    #         raise AlgorithmNameMismatchError('UPV', config_dict['algorithm_name'])
        
    #     self.gamma = config_dict['gamma']
    #     self.delta = config_dict['delta']
    #     self.z_threshold = config_dict['z_threshold']
    #     self.prefix_length = config_dict['prefix_length']
    #     self.bit_number = config_dict['bit_number']
    #     self.sigma = config_dict['sigma']
    #     self.default_top_k = config_dict['default_top_k']
    #     self.generator_model_name = config_dict['generator_model_name']
    #     self.detector_model_name = config_dict['detector_model_name']
    #     self.detect_mode = config_dict['detect_mode']

    #     # Validate detect mode
    #     if self.detect_mode not in ['key', 'network']:
    #         raise InvalidDetectModeError(self.detect_mode)

    #     self.generation_model = transformers_config.model
    #     self.generation_tokenizer = transformers_config.tokenizer
    #     self.vocab_size = transformers_config.vocab_size
    #     self.device = transformers_config.device
    #     self.gen_kwargs = transformers_config.gen_kwargs
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.gamma = self.config_dict['gamma']
        self.delta = self.config_dict['delta']
        self.z_threshold = self.config_dict['z_threshold']
        self.prefix_length = self.config_dict['prefix_length']
        self.bit_number = self.config_dict['bit_number']
        self.sigma = self.config_dict['sigma']
        self.default_top_k = self.config_dict['default_top_k']
        self.generator_model_name = self.config_dict['generator_model_name']
        self.detector_model_name = self.config_dict['detector_model_name']
        self.detect_mode = self.config_dict['detect_mode']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'UPV'


class UPVUtils:
    """Utility class for UPV algorithm, contains helper functions."""

    def __init__(self, config: UPVConfig, *args, **kwargs) -> None:
        """
            Initialize the UPV utility class.

            Parameters:
                config (UPVConfig): Configuration for the UPV algorithm.
        """
        self.config = config
        self.generator_model = self._get_generator_model(self.config.bit_number, self.config.prefix_length + 1).to(self.config.device)
        self.detector_model = self._get_detector_model(self.config.bit_number).to(self.config.device)
        self.cache = {}
        self.top_k = self.config.gen_kwargs.get('top_k', self.config.default_top_k)
        self.num_beams = self.config.gen_kwargs.get('num_beams', None)  

    def _get_generator_model(self, input_dim: int, window_size: int) -> UPVGenerator:
        """Load the generator model from the specified file."""
        model = UPVGenerator(input_dim, window_size)
        model.load_state_dict(torch.load(self.config.generator_model_name))
        return model

    def _get_detector_model(self, bit_number: int) -> UPVDetector:
        """Load the detector model from the specified file."""
        model = UPVDetector(bit_number)
        model.load_state_dict(torch.load(self.config.detector_model_name))
        return model

    def _get_predictions_from_generator(self, input_x: torch.Tensor) -> bool:
        """Get predictions from the generator model."""
        with torch.no_grad():
            output = self.generator_model(input_x)
            output = (output > 0.5).bool().item()
        return output
    
    def int_to_bin_list(self, n: int, length=8) -> list[int]:
        """Convert an integer to a binary list of specified length."""
        bin_str = format(n, 'b')[:length].zfill(length)
        return [int(b) for b in bin_str]
    
    def _select_candidates(self, scores: torch.Tensor) -> torch.Tensor:
        """Select candidate tokens based on the scores."""
        if self.num_beams is not None:
            threshold_score = torch.topk(scores, self.num_beams, largest=True, sorted=False)[0][-1]
            return (scores >= (threshold_score - self.config.delta)).nonzero(as_tuple=True)[0]
        else:
            return torch.topk(scores, self.top_k, largest=True, sorted=False).indices
    
    def get_greenlist_ids(self, input_ids: torch.Tensor, scores: torch.Tensor) -> list[int]:
        """Get greenlist ids for the input_ids."""
        greenlist_ids = []
        candidate_tokens = self._select_candidates(scores)
        
        # Ensure input_ids is a list for concatenation
        input_ids_list = input_ids.tolist() if isinstance(input_ids, torch.Tensor) else input_ids

        for v in candidate_tokens:
            # Now safely concatenate lists
            pair = input_ids_list[-self.config.prefix_length:] + [v.item()] if self.config.prefix_length > 0 else [v.item()]
            merged_tuple = tuple(pair)

            if merged_tuple in self.cache:
                result = self.cache[merged_tuple]
            else:
                bin_list = [self.int_to_bin_list(num, self.config.bit_number) for num in pair]
                result = self._get_predictions_from_generator(torch.tensor(bin_list, device=self.config.device).float().unsqueeze(0))
                self.cache[merged_tuple] = result
            if result:
                greenlist_ids.append(int(v))

        return greenlist_ids
    
    def _judge_green(self, input_ids: torch.Tensor, current_number: int) -> bool:
        """Judge if the current token is green based on previous tokens."""

        # Get the last 'prefix_length' items from input_ids
        last_nums = input_ids[-self.config.prefix_length:] if self.config.prefix_length > 0 else []
        # Append the current number to the list
        pair = list(last_nums) + [current_number]
        merged_tuple = tuple(pair)
        bin_list = [self.int_to_bin_list(num, self.config.bit_number) for num in pair]

        # load & update cache
        if merged_tuple in self.cache:
            result = self.cache[merged_tuple]
        else:
            result = self._get_predictions_from_generator(torch.tensor(bin_list, device=self.config.device).float().unsqueeze(0))
            self.cache[merged_tuple] = result

        return result
    
    def green_token_mask_and_stats(self, input_ids: torch.Tensor) -> tuple[list[bool], int, float]:
        """Get green token mask and statistics for the input_ids."""

        # Initialize a list with None for the prefix tokens which are not scored
        mask_list = [None] * self.config.prefix_length

        # Count of green tokens, initialized to zero
        green_token_count = 0

        # Iterate over each token in the input_ids starting from prefix_length
        for idx in range(self.config.prefix_length, len(input_ids)):
            # Get the current token
            curr_token = input_ids[idx]

            # Judge if the current token is green based on previous tokens
            if self._judge_green(input_ids[:idx], curr_token):
                mask_list.append(True)  # Mark this token as green
                green_token_count += 1  # Increment the green token counter
            else:
                mask_list.append(False)  # Mark this token as not green

        # Compute the number of tokens that were evaluated for green status
        num_tokens_scored = len(input_ids) - self.config.prefix_length

        # Calculate the z-score for the number of green tokens
        z_score = self._compute_z_score(green_token_count, num_tokens_scored)

        # Return the mask list, count of green tokens, and the z-score
        return mask_list, green_token_count, z_score

    def _compute_z_score(self, observed_count: int, T: int) -> float:
        """Compute z-score for the given observed count and total tokens."""
        expected_count = self.config.gamma
        numer = observed_count - expected_count * T
        denom = sqrt(T * expected_count * (1 - expected_count) + self.config.sigma * self.config.sigma * T)
        z = numer / denom
        return z


class UPVLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for UPV algorithm, process logits to add watermark."""

    def __init__(self, config: UPVConfig, utils: UPVUtils, *args, **kwargs):
        """
            Initialize the UPV logits processor.

            Parameters:
                config (UPVConfig): Configuration for the UPV algorithm.
                utils (UPVUtils): Utility class for the UPV algorithm.
        """
        self.config = config
        self.utils = utils

    def _bias_greenlist_logits(self, scores: torch.Tensor, greenlist_mask: torch.Tensor, greenlist_bias: float) -> torch.Tensor:
        """Bias the logits for the greenlist tokens."""
        scores[greenlist_mask] = scores[greenlist_mask] + greenlist_bias
        return scores

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process the logits to add watermark."""
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores
        
        green_tokens_mask = torch.zeros_like(scores)
        for b_idx in range(input_ids.shape[0]):
            greenlist_ids = self.utils.get_greenlist_ids(input_ids[b_idx], scores=scores[b_idx])
            green_tokens_mask[b_idx][greenlist_ids] = 1 
        green_tokens_mask = green_tokens_mask.bool()

        scores = self._bias_greenlist_logits(scores=scores, greenlist_mask=green_tokens_mask, greenlist_bias=self.config.delta)

        return scores


class UPV(BaseWatermark):
    """Top-level class for UPV algorithm."""

    def __init__(self, algorithm_config: str | UPVConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the UPV algorithm.

            Parameters:
                algorithm_config (str | UPVConfig): Path to the algorithm configuration file or UPVConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = UPVConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, UPVConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a UPVConfig instance")
        
        self.utils = UPVUtils(self.config)
        self.logits_processor = UPVLogitsProcessor(self.config, self.utils)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text based on the prompt."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        return watermarked_text

    def _detect_watermark_network_mode(self, encoded_text: torch.Tensor) -> tuple[bool, float]:
        """ Detect watermark using the network mode. """
        # Convert input IDs to binary sequence
        inputs_bin = [self.utils.int_to_bin_list(token_id, self.config.bit_number) for token_id in encoded_text]
        inputs_bin = torch.tensor(inputs_bin, device=self.config.device)

        # Run the model on the input binary sequence
        outputs = self.utils.detector_model(inputs_bin.unsqueeze(dim=0).float())
        outputs = outputs.reshape([-1])
        predicted = (outputs.data > 0.5).int()

        # Determine watermark presence based on predictions
        is_watermarked = (predicted == 1).sum().item() > 0

        # z_score is not applicable in network mode
        return is_watermarked, None  
    
    def detect_watermark(self, text: str, return_dict: bool = True) -> tuple[bool, float]:
        """Detect watermark in the given text."""
        
        # Encode the text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0]

        # Check the mode and perform detection accordingly
        if self.config.detect_mode == 'key':
            _, _, z_score = self.utils.green_token_mask_and_stats(encoded_text)
            # Determine if the z_score indicates a watermark
            is_watermarked = z_score > self.config.z_threshold
        else:
            is_watermarked, z_score = self._detect_watermark_network_mode(encoded_text)
        
        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)

    def get_data_for_visualization(self, text: str) -> DataForVisualization:
        """Get data for visualization."""

        # Encode the text using the specified tokenizer configuration
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0]
        
        # Get the mask indicating green tokens
        mask, _, _ = self.utils.green_token_mask_and_stats(encoded_text)
        
        # Create a list of flags from the mask where -1 = None, 1 = True (green), 0 = False (not green)
        highlight_values = [-1 if m is None else 1 if m else 0 for m in mask]
        
        # Decode each token ID to its corresponding token string
        decoded_tokens = [self.config.generation_tokenizer.decode(token_id.item()) for token_id in encoded_text]
        
        # Return the data for visualization
        return DataForVisualization(decoded_tokens, highlight_values)



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ==================================================
# exp_gumbel.py
# Description: Implementation of EXPGumbel algorithm
# ==================================================

import scipy
import torch
from math import log
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from transformers import LogitsProcessor
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from visualize.data_for_visualization import DataForVisualization

class EXPGumbelConfig(BaseConfig):
    """Config class for EXPGumbel algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.prefix_length = self.config_dict['prefix_length']
        self.eps = self.config_dict['eps']
        self.threshold = self.config_dict['threshold']
        self.sequence_length = self.config_dict['sequence_length']
        self.temperature = self.config_dict['temperature']
        self.seed = self.config_dict['seed']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'EXPGumbel'


class EXPGumbelUtils:
    """Utility class for EXPGumbel algorithm, contains helper functions."""

    def __init__(self, config: EXPGumbelConfig, *args, **kwargs) -> None:
        """
            Initialize the EXPGumbel utility class.

            Parameters:
                config (EXPGumbelConfig): Configuration for the EXPGumbel algorithm.
        """
        self.config = config
        self.generator = torch.Generator().manual_seed(self.config.seed)
        self.uniform = torch.clamp(torch.rand((self.config.vocab_size * self.config.prefix_length, self.config.vocab_size), 
                                         generator=self.generator, dtype=torch.float32), min=self.config.eps)
        self.gumbel = (-torch.log(torch.clamp(-torch.log(self.uniform), min=self.config.eps))).to(self.config.device)
    
    def _value_transformation(self, value):
        """Transform the value to a range between 0 and 1."""
        return value/(value + 1)


class EXPGumbelLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for EXPGumbel algorithm, process logits to add watermark."""

    def __init__(self, config: EXPGumbelConfig, utils: EXPGumbelUtils, *args, **kwargs) -> None:
        """
            Initialize the EXPGumbel logits processor.

            Parameters:
                config (EXPGumbelConfig): Configuration for the KGW algorithm.
                utils (EXPGumbelUtils): Utility class for the KGW algorithm.
        """
        self.config = config
        self.utils = utils

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores
        prev_token = torch.sum(input_ids[:, -self.config.prefix_length:], dim=-1)  # (batch_size,)
        gumbel = self.utils.gumbel[prev_token]  # (batch_size, vocab_size)
        return scores[..., :gumbel.shape[-1]] / self.config.temperature + gumbel


class EXPGumbel(BaseWatermark):
    """Top-level class for the EXPGumbel algorithm."""

    def __init__(self, algorithm_config: str | EXPGumbelConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the EXPGumbel algorithm.

            Parameters:
                algorithm_config (str | EXPGumbelConfig): Path to the algorithm configuration file or EXPGumbelConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = EXPGumbelConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, EXPGumbelConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a EXPGumbelConfig instance")
        
        self.utils = EXPGumbelUtils(self.config)
        self.logits_processor = EXPGumbelLogitsProcessor(self.config, self.utils)
    
    def watermark_logits_argmax(self, input_ids: torch.LongTensor, logits: torch.FloatTensor) -> torch.LongTensor:
        """
        Applies watermarking to the last token's logits and returns the argmax for that token.
        Returns tensor of shape (batch,), where each element is the index of the selected token.
        """
        
        # Get the logits for the last token
        last_logits = logits[:, -1, :]  # (batch, vocab_size)
        
        # Get the argmax of the logits
        last_token = torch.argmax(last_logits, dim=-1).unsqueeze(-1)  # (batch,)
        return last_token

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text using the EXPGumbel algorithm."""

        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer.encode(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)

        # Initialize
        inputs = encoded_prompt
        attn = torch.ones_like(encoded_prompt)
        past = None

        # Generate tokens
        for i in range(self.config.sequence_length):
            with torch.no_grad():
                if past:
                    output = self.config.generation_model(inputs[:,-1:], past_key_values=past, attention_mask=attn)
                    output_gumbel = self.logits_processor(input_ids=inputs, scores=output.logits)
                else:
                    output = self.config.generation_model(inputs)
                    output_gumbel = self.logits_processor(input_ids=inputs, scores=output.logits)

            # Sample token
            token = self.watermark_logits_argmax(inputs, output_gumbel)

            # Update past
            past = output.past_key_values

            # Update inputs
            inputs = torch.cat([inputs, token], dim=-1)

            # Update attention mask
            attn = torch.cat([attn, attn.new_ones((attn.shape[0], 1))], dim=-1)

        watermarked_tokens = inputs[0].detach().cpu()
        watermarked_text = self.config.generation_tokenizer.decode(watermarked_tokens, skip_special_tokens=True)

        return watermarked_text

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs) -> dict:
        """Detect watermark in the text."""

        # Encode the text into tokens using the configured tokenizer
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False)[0]

        seq_len = len(encoded_text)
        score = 0
        for i in range(self.config.prefix_length, seq_len):
            prev_tokens_sum = torch.sum(encoded_text[i - self.config.prefix_length:i], dim=-1)
            token = encoded_text[i]
            u = self.utils.uniform[prev_tokens_sum, token]
            score += log(1 / (1 - u))
        
        p_value = scipy.stats.gamma.sf(score, seq_len - self.config.prefix_length, loc=0, scale=1)
        
        # Determine if the computed score exceeds the threshold for watermarking
        is_watermarked = bool(p_value < self.config.threshold)

        # Return results based on the `return_dict` flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": p_value}
        else:
            return (is_watermarked, p_value)
    
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> DataForVisualization:
        """Get data for visualization."""
        
        # Encode the text into tokens using the configured tokenizer
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False)[0]

        # Initialize the list of values with None for the prefix length
        highlight_values = [None] * self.config.prefix_length

        # Calculate the value for each token beyond the prefix
        for i in range(self.config.prefix_length, len(encoded_text)):
            # Seed the random number generator using the prefix of the encoded text
            prev_tokens_sum = torch.sum(encoded_text[i - self.config.prefix_length:i], dim=-1)
            token = encoded_text[i]
            u = self.utils.uniform[prev_tokens_sum, token]
            score = log(1 / (1 - u))
            highlight_values.append(self.utils._value_transformation(score))

        # Decode each token id to its corresponding token
        decoded_tokens = [self.config.generation_tokenizer.decode([token_id]) for token_id in encoded_text]

        return DataForVisualization(decoded_tokens, highlight_values)
            


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ===========================================================
# unbiased.py
# Description: Implementation of Unbiased Watermark algorithm
# ===========================================================

import torch
import hashlib
import numpy as np
import numpy.typing as npt
import random
from typing import Tuple, Union
import torch.nn.functional as F
from math import sqrt
from functools import partial
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization
from .strategy import DeltaStrategy, GammaStrategy

class UnbiasedConfig(BaseConfig):
    """Config class for Unbiased watermark algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.alpha = 0.5
        self.ignore_history_generation = bool(self.config_dict['ignore_history_generation'])
        self.ignore_history_detection = bool(self.config_dict['ignore_history_detection'])
        self.p_threshold = self.config_dict['p_threshold']
        self.prefix_length = self.config_dict['prefix_length']  
        self.type = self.config_dict['type']
        self.hash_key = random.getrandbits(1024).to_bytes(128, "big")
        self.n_grid = self.config_dict['n_grid']
        
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'Unbiased'

class UnbiasedUtils:
    """Utility class for Unbiased watermark algorithm, contains helper functions."""

    def __init__(self, config: UnbiasedConfig, *args, **kwargs) -> None:
        """
            Initialize the Unbiased utility class.

            Parameters:
                config (UnbiasedConfig): Configuration for the unbiased algorithm.
        """
        self.config = config
        self.rng = torch.Generator(device=self.config.device)
        self.cc_history = set()
        self.state_indicator = 0 # 0 for generation, 1 for detection and visualization
        self.strategy = {
            "delta": DeltaStrategy(),
            "gamma": GammaStrategy()
        }[self.config.type]
        

    def _get_rng_seed(self, context_code: any) -> int:
        """Get the random seed from the given context code and private key."""
        if (
            (not self.config.ignore_history_generation and self.state_indicator == 0) or 
            (not self.config.ignore_history_detection and self.state_indicator == 1)
        ):
            self.cc_history.add(context_code)
            
        m = hashlib.sha256()
        m.update(context_code)
        m.update(self.config.hash_key)
        
        full_hash = m.digest()
        seed = int.from_bytes(full_hash, "big") % (2**32 - 1)
        return seed
    
    def _extract_context_code(self, context: torch.LongTensor) -> bytes:
        """Extract context code from the given context."""
        if self.config.prefix_length == 0:
            return context.detach().cpu().numpy().tobytes()
        else:
            return context[-self.config.prefix_length : ].detach().cpu().numpy().tobytes()
    
    def get_seed_for_cipher(self, input_ids: torch.LongTensor) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Get the mask and seeds for the cipher."""
        batch_size = input_ids.size(0)
        context_codes = [
            self._extract_context_code(input_ids[i]) for i in range(batch_size)
        ]

        mask, seeds = zip(
            *[
                (context_code in self.cc_history, self._get_rng_seed(context_code))
                for context_code in context_codes 
            ]
        )
        
        return mask, seeds
    
    def _apply_watermark(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Apply watermark to the scores."""
        mask, seeds = self.get_seed_for_cipher(input_ids)
        
        rng = [
            torch.Generator(device=scores.device).manual_seed(seed) for seed in seeds
        ]
        mask = torch.tensor(mask, device=scores.device)
        shuffle = self.strategy.from_random(
            rng, scores.size(1)
        )

        reweighted_scores = self.strategy.reweight_logits(shuffle, scores)
        
        return mask, reweighted_scores
    
    def _safe_minus(self, q_logits: torch.FloatTensor, p_logits: torch.FloatTensor) -> torch.FloatTensor:
        """Safe minus operation to avoid numerical instability."""
        llr = q_logits - p_logits
        llr.nan_to_num_(nan=0.0)
        return llr
    
    def _from_grid(self, dist_ps: npt.ArrayLike, dist_qs: npt.ArrayLike):
        """Generate batch query from grid search."""
        dist_ps = np.array(dist_ps)
        dist_qs = np.array(dist_qs)
        assert dist_ps.ndim == 1
        assert dist_qs.ndim == 1
        assert np.all(dist_ps >= 0) and np.all(dist_qs >= 0)
        with np.errstate(divide="ignore"):
            dist_p_logs = np.log(dist_ps)
            dist_q_logs = np.log(dist_qs)
        dist_p_logs.sort()
        dist_q_logs.sort()
        batch_query = [(d_p_l, d_q_l) for d_p_l in dist_p_logs for d_q_l in dist_q_logs]
        return batch_query
    
    def _get_max_llr(
        self, # shape = (..., vocab_size)
        p_logits: torch.FloatTensor,
        q_logits: torch.FloatTensor,
        batch_query: list[tuple[float, float]],
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:
        """Get the max llr"""
        # shape = (..., vocab_size)
        llr = self._safe_minus(q_logits, p_logits)
        # shape = (..., vocab_size)
        try:
            sort_index = torch.argsort(llr, dim=-1, descending=True)
        except torch.cuda.OutOfMemoryError as e:
            # use cpu instead
            sort_index = torch.argsort(llr.cpu(), dim=-1, descending=True).to(llr.device)
        del llr

        p_logits = p_logits.gather(-1, sort_index)
        q_logits = q_logits.gather(-1, sort_index)
        del sort_index

        # shape = (..., vocab_size)
        llr = self._safe_minus(q_logits, p_logits)

        # shape = (..., vocab_size)
        sum_q_logits = torch.logcumsumexp(q_logits, dim=-1)
        sum_p_logits = torch.logcumsumexp(p_logits, dim=-1)
        del q_logits
        del p_logits

        max_llrs = []
        for dist_p_log, dist_q_log in batch_query:
            # shape = (..., vocab_size)
            modified_q_logits = torch.where(
                sum_q_logits <= dist_q_log,
                torch.tensor(
                    float("-inf"), device=sum_q_logits.device, dtype=sum_q_logits.dtype
                ),
                sum_q_logits + torch.log(-torch.expm1(dist_q_log - sum_q_logits)),
            )
            modified_p_logits = torch.logaddexp(
                sum_p_logits,
                torch.tensor(
                    dist_p_log, device=sum_p_logits.device, dtype=sum_p_logits.dtype
                ),
            )

            # shape = (..., vocab_size)
            modified_llr = self._safe_minus(modified_q_logits, modified_p_logits)
            del modified_p_logits
            del modified_q_logits

            # pad left modified_llr with -inf
            # shape = (..., vocab_size+1)
            modified_llr = F.pad(modified_llr, (1, 0), value=float("-inf"))
            # get index by argmax
            # shape = (..., )
            cut_index = torch.where(
                torch.any(llr < modified_llr[..., :-1], dim=-1),
                torch.argmax((llr < modified_llr[..., :-1]).to(torch.int), dim=-1),
                torch.tensor(modified_llr.shape[-1] - 1, device=modified_llr.device),
            )
            # shape = (..., 1)
            max_llrs.append(modified_llr.gather(-1, cut_index.unsqueeze(-1)))
        # shape = (..., query_size)
        max_llr = torch.cat(max_llrs, dim=-1)
        del max_llrs
        return max_llr
    
    @torch.no_grad()
    def _score_llr(
        self, 
        p_logits: torch.FloatTensor, 
        q_logits: torch.FloatTensor,
        batch_query: list[tuple[float, float]],
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:
        """
            return (llr, max_llr, min_llr)
            llr: [batch_size, seq_len, vocab_size]
            max_llr: [batch_size, seq_len, query_size]
            min_llr: [batch_size, seq_len, query_size]
        """
        q_logits = F.log_softmax(q_logits, dim=-1)
        p_logits = F.log_softmax(p_logits, dim=-1)

        max_llr = self._get_max_llr(p_logits, q_logits, batch_query)
        min_llr = -self._get_max_llr(
            q_logits, p_logits, [(q, p) for p, q in batch_query]
        )
        trivial_pos = max_llr < min_llr
        max_llr = torch.where(
            trivial_pos, torch.tensor(0.0, device=max_llr.device), max_llr
        )
        min_llr = torch.where(
            trivial_pos, torch.tensor(0.0, device=min_llr.device), min_llr
        )

        llr = self._safe_minus(q_logits, p_logits)
        return llr, max_llr, min_llr
    
    def value_transformation(self, value: float) -> float:
        """Transform value to range [0, 1]."""
        return value/(value + 1)
    
    def score_sequence(self, text: str) -> tuple[float, list[int]]:
        """Score the input_ids and return z_score and green_token_flags."""
        # Step 1: Get dist for grid-search
        n = self.config.n_grid
        dist = [float(i) / n for i in range(n + 1)]
        batch_query = self._from_grid([0.0], dist)
        
        # Step 2: Get the original logits & modified logits
        inputs = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False).to(self.config.device)
        ## Store the full input_ids for later use
        full_input_ids = inputs["input_ids"]
        ## Ignore new token logits
        input_ids = inputs["input_ids"][..., :-1].to(self.config.generation_model.device)
        attention_mask = inputs["attention_mask"][..., :-1].to(self.config.generation_model.device)
        # Ignore prefix token's score
        labels = inputs["input_ids"][..., self.config.prefix_length:].to(self.config.generation_model.device)
        
        outputs = self.config.generation_model(input_ids=input_ids, attention_mask=attention_mask, use_cache=False)
        logits = outputs.logits
        old_logits = torch.clone(logits)
        new_logits = torch.clone(logits)
        
        for i in range(self.config.prefix_length - 1, logits.size(1) - 1): # Skip prefix tokens & logits for new token
            pre_input_ids = input_ids[:, :i+1]
            # Original logits 
            t = logits[:, i]
            # Modified logits
            mask, reweighted_scores = self._apply_watermark(pre_input_ids, t)
            
            old_logits[:, i] = t
            if self.config.ignore_history_detection:
                new_logits[:, i] = reweighted_scores
            else:
                new_logits[:, i] = torch.where(mask[:, None], t, reweighted_scores)
                
        old_logits = old_logits[:, self.config.prefix_length - 1:]
        new_logits = new_logits[:, self.config.prefix_length - 1:]
        
        # Step 3: Get llr, max_llr, min_llr
        llr, max_llr, min_llr = self._score_llr(old_logits, new_logits, batch_query)
        
        # Step 4: Extract & Clamp llr for text's token ids
        unclipped_scores = torch.gather(llr, -1, labels.unsqueeze(-1)).squeeze(-1)
        ## shape = (batch_size, seq_len - prefix_length, query_size)
        scores = torch.clamp(unclipped_scores.unsqueeze(-1), min_llr, max_llr)
        scores = np.array(scores[0].cpu())
        labels = np.array(labels[0].cpu())
        
        # Step 5: Choose the best index for grid search
        sum_scores = np.sum(scores, axis=0)
        best_index = np.argmax(sum_scores)
        final_score = sum_scores[best_index]
        
        # Step 6: Calculate probability & token_list for visualization
        ## get the best score for each token (dim = 1)
        best_scores = scores[:, best_index]
        
        ## Create highlight_values array matching the FULL input length
        highlight_values = [None] * len(full_input_ids[0])
        
        ## Place the scores at their correct positions, starting from prefix_length
        for i, score in enumerate(best_scores):
            position = i + self.config.prefix_length
            if position < len(highlight_values):
                highlight_values[position] = self.value_transformation(score)
        
        ## Theorem 9: p <= A * e^(-t)
        p_val = n * np.exp(-final_score)
        
        return p_val, highlight_values

class UnbiasedLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for DiP algorithm, process logits to add watermark."""

    def __init__(self, config: UnbiasedConfig, utils: UnbiasedUtils, *args, **kwargs) -> None:
        """
            Initialize the Unbiased logits processor.

            Parameters:
                config (UnbiasedConfig): Configuration for the Unbiased algorithm.
                utils (UnbiasedUtils): Utility class for the Unbiased algorithm.
        """
        self.config = config
        self.utils = utils

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process logits to add watermark."""
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores
        
        mask, reweighted_scores = self.utils._apply_watermark(input_ids, scores)
        
        if self.config.ignore_history_generation:
            return reweighted_scores
        else:
            return torch.where(mask[:, None], scores, reweighted_scores)
    

class UnbiasedWatermark(BaseWatermark):
    """Top-level class for Unbiased algorithm."""

    def __init__(self, algorithm_config: str | UnbiasedConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the Unbiased algorithm.

            Parameters:
                algorithm_config (str | UnbiasedConfig): Path to the algorithm configuration file or UnbiasedConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = UnbiasedConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, UnbiasedConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a UnbiasedConfig instance")
        
        self.utils = UnbiasedUtils(self.config)
        self.logits_processor = UnbiasedLogitsProcessor(self.config, self.utils)
    
    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text."""

        # Set the state indicator to 0 for generation
        self.utils.state_indicator = 0
        
        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # Generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # Decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        # Clear the history
        self.utils.cc_history.clear()
        return watermarked_text
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""
        
        # Set the state indicator to 1 for detection
        self.utils.state_indicator = 1
        
        # Compute z-score using a utility method
        p_val, _ = self.utils.score_sequence(text)
        
        # Determine if the z-score indicates a watermark
        is_watermarked = p_val < self.config.p_threshold

        # Clear the history
        self.utils.cc_history.clear()

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": p_val}
        else:
            return (is_watermarked, p_val)
        
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
        """Get data for visualization."""
        
        # Set the state indicator to 1 for visualization
        self.utils.state_indicator = 1
        
        # Compute the highlight values
        _, highlight_values = self.utils.score_sequence(text)
        
        # Encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
        # decode single tokens
        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)
        
        # Clear the history
        self.utils.cc_history.clear()
        
        return DataForVisualization(decoded_tokens, highlight_values)



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# transformers_config.py
# Description: Configuration for transformers
# ============================================

class TransformersConfig:
    """Configuration class for transformers."""

    def __init__(self, model, tokenizer, vocab_size=None, device='cuda', *args, **kwargs):
        """
            Initialize the transformers configuration.

            Parameters:
                model (object): The model object.
                tokenizer (object): The tokenizer object.
                vocab_size (int): The vocabulary size.
                device (str): The device to use.
                kwargs: Additional keyword arguments.
        """
        self.device = device
        self.model = model
        self.tokenizer = tokenizer
        self.vocab_size = len(tokenizer) if vocab_size is None else vocab_size
        self.gen_kwargs = {}
        self.gen_kwargs.update(kwargs)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ==============================================
# morphmark.py
# Description: Implementation of MorphMark algorithm
# ==============================================

import os
import time
import numpy as np
import torch
from math import sqrt
from functools import partial
from ..base import BaseConfig, BaseWatermark
from utils.utils import load_config_file
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization


class MorphMarkConfig(BaseConfig):
    """Config class for MorphMark algorithm, load config file and initialize parameters."""
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""

        self.gamma = self.config_dict['gamma']  # Proportion of the green table

        self.type = self.config_dict['type']  # Type of MorphMark, three options: linear, exp, log
        self.k_linear = self.config_dict['k_linear']
        self.k_exp = self.config_dict['k_exp']
        self.k_log = self.config_dict['k_log']
        self.p_0 = self.config_dict['p_0']

        self.ewd = self.config_dict['ewd']  # Whether to use ewd as the watermark detection method
        self.delta = self.config_dict['delta_ewd']  # Parameter for ewd

        self.hash_key = self.config_dict['hash_key']
        self.z_threshold = self.config_dict['z_threshold']
        self.prefix_length = self.config_dict['prefix_length']
        self.f_scheme = self.config_dict['f_scheme']
        self.window_scheme = self.config_dict['window_scheme']

    @property
    def algorithm_name(self) -> str:
        """Return algorithm name."""
        return 'MorphMark'

class MorphMarkUtils:
    """Utility class for MorphMark algorithm, contains helper functions."""

    def __init__(self, config: MorphMarkConfig, *args, **kwargs) -> None:
        """
            Initialize the MorphMark utility class.

            Parameters:
                config (MorphMarkConfig): Configuration for the MorphMark algorithm.
        """
        self.config = config
        self.rng = torch.Generator(device=self.config.device)
        self.rng.manual_seed(self.config.hash_key)
        self.prf = torch.randperm(self.config.vocab_size, device=self.config.device, generator=self.rng)
        self.f_scheme_map = {"time": self._f_time, "additive": self._f_additive, "skip": self._f_skip, "min": self._f_min}
        self.window_scheme_map = {"left": self._get_greenlist_ids_left, "self": self._get_greenlist_ids_self}

        alpha = torch.exp(torch.tensor(self.config.delta)).item()
        self.z_value = ((1 - self.config.gamma) * (alpha - 1))/(1 - self.config.gamma + (alpha * self.config.gamma))
            
    def _f(self, input_ids: torch.LongTensor) -> int:
        """Get the previous token."""
        return int(self.f_scheme_map[self.config.f_scheme](input_ids))

    def _f_time(self, input_ids: torch.LongTensor):
        """Get the previous token time."""
        time_result = 1
        for i in range(0, self.config.prefix_length):
            time_result *= input_ids[-1 - i].item()
        return self.prf[time_result % self.config.vocab_size]

    def _f_additive(self, input_ids: torch.LongTensor):
        """Get the previous token additive."""
        additive_result = 0
        for i in range(0, self.config.prefix_length):
            additive_result += input_ids[-1 - i].item()
        return self.prf[additive_result % self.config.vocab_size]

    def _f_skip(self, input_ids: torch.LongTensor):
        """Get the previous token skip."""
        return self.prf[input_ids[- self.config.prefix_length].item()]

    def _f_min(self, input_ids: torch.LongTensor):
        """Get the previous token min."""
        return min(self.prf[input_ids[-1 - i].item()] for i in range(0, self.config.prefix_length))

    def get_greenlist_ids(self, input_ids: torch.LongTensor, is_detect=False) -> list[int]:
        """Get greenlist ids for the input_ids."""
        return self.window_scheme_map[self.config.window_scheme](input_ids, is_detect)

    def _seed_rng(self, input_ids: torch.LongTensor) -> None:
        """Seed the random number generator with the last token in the input_ids."""
        time_result = 1
        for i in range(0, self.config.prefix_length):
            time_result *= input_ids[-1 - i].item()
        prev_token = time_result % self.config.vocab_size
        self.rng.manual_seed(self.config.hash_key * prev_token)
        return
        
    def _get_greenlist_ids_left(self, input_ids: torch.LongTensor, is_detect=False) -> list[int]:
        """Get greenlist ids for the input_ids via leftHash scheme."""
        if self.config.ewd:
            self._seed_rng(input_ids)
        else:
            self.rng.manual_seed((self.config.hash_key * self._f(input_ids)) % self.config.vocab_size)
        greenlist_size = int(self.config.vocab_size * self.config.gamma)
        vocab_permutation = torch.randperm(self.config.vocab_size, device=input_ids.device, generator=self.rng)
        greenlist_ids = vocab_permutation[:greenlist_size]

        return greenlist_ids

    def _get_greenlist_ids_self(self, input_ids: torch.LongTensor) -> list[int]:
        """Not Implement."""
        raise "Not Implemented 'def _get_greenlist_ids_self'"

    def calculate_entropy(self, model, tokenized_text) -> list[float]:
        """Calculate the entropy of the tokenized text using the model."""
        with torch.no_grad():
            output = model(torch.unsqueeze(tokenized_text, 0), return_dict=True)
            probs = torch.softmax(output.logits, dim=-1)
            denoms = 1+(self.z_value * probs)
            renormed_probs = probs / denoms
            sum_renormed_probs = renormed_probs.sum(dim=-1)
            entropy=sum_renormed_probs[0].cpu().tolist()
            entropy.insert(0, -10000.0)
            return entropy[:-1]
    
    def _get_weight_from_entropy(self, entropy_list: list[float]) -> list:
        """Compute the weights from the entropy list."""

        # Convert the entropy list to a tensor
        entropy_tensor = torch.tensor(entropy_list)

        # Compute the minimum entropy from the elements beyond the prefix length
        min_entropy = torch.min(entropy_tensor[self.config.prefix_length:])

        # Subtract this minimum entropy from all entropy values
        adjusted_entropies = entropy_tensor - min_entropy

        # Create a list where the prefix part is filled with -1, and the rest with adjusted entropies
        weights = [-1] * self.config.prefix_length + adjusted_entropies[self.config.prefix_length:].tolist()

        return weights

    def _compute_z_score(self, observed_count: int , T: int) -> float:
        """Compute z-score for the given observed count and total tokens."""
        expected_count = self.config.gamma
        numer = observed_count - expected_count * T
        denom = sqrt(T * expected_count * (1 - expected_count))
        z = numer / denom
        return z

    def _compute_z_score_ewd(self, observed_count: float, weight: list) -> float:
        """Compute the z-score for the given observed count and weight."""

        weight_tensor = torch.tensor(weight, dtype=torch.float)
        expected_count = self.config.gamma
        numer = observed_count - expected_count * torch.sum(weight_tensor)
        denom = torch.sqrt(torch.sum(torch.square(weight_tensor)) * expected_count * (1 - expected_count))
        z = numer / denom
        return z.item()

    def score_sequence(self, input_ids: torch.Tensor, entropy_list=None):
        """Score the input_ids and return z_score and green
        _token_flags."""
        num_tokens_scored = len(input_ids) - self.config.prefix_length
        if num_tokens_scored < 1:
            raise ValueError(
                (
                    f"Must have at least {1} token to score after "
                    f"the first min_prefix_len={self.config.prefix_length} tokens required by the seeding scheme."
                )
            )

        green_token_count = 0
        green_token_flags = [-1 for _ in range(self.config.prefix_length)]

        for idx in range(self.config.prefix_length, len(input_ids)):
            curr_token = input_ids[idx]
            greenlist_ids = self.get_greenlist_ids(input_ids[:idx], is_detect=True)
            if curr_token in greenlist_ids:
                green_token_count += 1
                green_token_flags.append(1)
            else:
                green_token_flags.append(0)

        # compute z_score
        if self.config.ewd:
            # calculate weights
            weights = self._get_weight_from_entropy(entropy_list)
            # sum up weights where green_token_flags = 1 to get green_token_count
            green_token_count = sum(weights[i] for i in range(len(green_token_flags)) if green_token_flags[i] == 1)
            z_score = self._compute_z_score_ewd(green_token_count, weights[self.config.prefix_length:])
            return z_score, green_token_flags, weights
        else:
            z_score = self._compute_z_score(green_token_count, num_tokens_scored)
            return z_score, green_token_flags


class MorphMarkLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for MorphMark algorithm, process logits to add watermark."""

    def __init__(self, config: MorphMarkConfig, utils: MorphMarkUtils, *args, **kwargs) -> None:
        """
            Initialize the MorphMark logits processor.

            Parameters:
                config (MorphMarkConfig): Configuration for the MorphMark algorithm.
                utils (MorphMarkUtils): Utility class for the MorphMark algorithm.
        """
        self.config = config
        self.utils = utils

    def _calc_greenlist_mask(self, scores: torch.FloatTensor, greenlist_token_ids: torch.LongTensor) -> torch.BoolTensor:
        """Calculate greenlist mask for the given scores and greenlist token ids."""
        green_tokens_mask = torch.zeros_like(scores)
        for b_idx in range(len(greenlist_token_ids)):
            green_tokens_mask[b_idx][greenlist_token_ids[b_idx]] = 1
        final_mask = green_tokens_mask.bool()
        return final_mask

    def _bias_greenlist_logits(self, scores: torch.Tensor, greenlist_mask: torch.Tensor) -> torch.Tensor:
        # scores = scores.to(torch.float32)  # due to float16 can not equal to 10^50
        probs = torch.softmax(scores / 1.0, dim=-1)

        P_G = probs[greenlist_mask].sum().item()

        if P_G < self.config.p_0:
            r = 0.0
        else:
            if self.config.type == "linear":
                r = self.config.k_linear * P_G
            elif self.config.type == "exp":
                r = np.exp(self.config.k_exp * P_G) - 1
            elif self.config.type == "log":
                r = np.log(self.config.k_log * P_G + 1)
            else:
                raise ValueError(f"{self.config.type} is not defined.")

        beta = r * (1 - P_G)
        weights = probs[greenlist_mask]
        normalized_weights = weights / weights.sum()
        probs[greenlist_mask] = probs[greenlist_mask] + normalized_weights * beta
        weights = probs[~greenlist_mask]
        normalized_weights = weights / weights.sum()
        probs[~greenlist_mask] = probs[~greenlist_mask] - normalized_weights * beta
        probs = torch.nan_to_num(probs, nan=0)
        probs = torch.clamp(probs, min=0)
        probs = probs / probs.sum(dim=-1, keepdim=True)
        scores = torch.log(probs) * 1.0

        return scores


    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process logits to add watermark."""
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores

        batched_greenlist_ids = [None for _ in range(input_ids.shape[0])]

        for b_idx in range(input_ids.shape[0]):
            greenlist_ids = self.utils.get_greenlist_ids(input_ids[b_idx], is_detect=False)
            batched_greenlist_ids[b_idx] = greenlist_ids

        green_tokens_mask = self._calc_greenlist_mask(scores=scores, greenlist_token_ids=batched_greenlist_ids)

        scores = self._bias_greenlist_logits(scores=scores, greenlist_mask=green_tokens_mask)
        return scores
    

class MorphMark(BaseWatermark):
    """Top-level class for MorphMark algorithm."""

    def __init__(self, algorithm_config: str | MorphMarkConfig, transformers_config: TransformersConfig | None = None, *args, ** kwargs) -> None:
        """
            Initialize the MorphMark algorithm.

            Parameters:
                algorithm_config (str): Path to the algorithm configuration file.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """

        if isinstance(algorithm_config, str):
            self.config = MorphMarkConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, MorphMarkConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a MorphMarkConfig instance")

        self.utils = MorphMarkUtils(self.config)
        self.logits_processor = MorphMarkLogitsProcessor(self.config, self.utils)
    
    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # Generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # Decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        return watermarked_text
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""

        # Encode the text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)

        # calculate entropy
        if self.config.ewd:
            entropy_list = self.utils.calculate_entropy(self.config.generation_model, encoded_text)

        # Compute z_score using a utility method
        if self.config.ewd:
            z_score, _, _ = self.utils.score_sequence(encoded_text, entropy_list)
        else:
            z_score, _ = self.utils.score_sequence(encoded_text)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)
        
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
        """Get data for visualization."""
        
        # Encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
        # Compute z-score and highlight values
        z_score, highlight_values = self.utils.score_sequence(encoded_text)
        
        # decode single tokens

        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)
        
        return DataForVisualization(decoded_tokens, highlight_values)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# transform_model.py
# Description: Transform model for X-SIR algorithm
# ============================================

import torch.nn as nn

class ResidualBlock(nn.Module):
    def __init__(self, dim):
        super(ResidualBlock, self).__init__()
        self.fc = nn.Linear(dim, dim)
        self.relu = nn.ReLU()

    def forward(self, x):
        out = self.fc(x)
        out = self.relu(out)
        out = out + x 
        return out

class TransformModel(nn.Module):
    def __init__(self, num_layers=4, input_dim=1024, hidden_dim=500, output_dim=300):
        super(TransformModel, self).__init__()
        
        self.layers = nn.ModuleList()
        
        self.layers.append(nn.Linear(input_dim, hidden_dim))
        
        for _ in range(num_layers - 2):
            self.layers.append(ResidualBlock(hidden_dim))

        self.layers.append(nn.Linear(hidden_dim, output_dim))

    def forward(self, x):
        layer_dtype = self.layers[0].weight.dtype
        x = x.to(layer_dtype)
        for i in range(len(self.layers)):
            x = self.layers[i](x)
        return x


import torch
import torch.nn.functional as F
from typing import Union

class WatermarkStrategy:
    """Base class for watermark strategies."""
    def from_random(self, rng: Union[torch.Generator, list[torch.Generator]], vocab_size: int) -> torch.LongTensor:
        raise NotImplementedError
        
    def reweight_logits(self, shuffle: torch.LongTensor, p_logits: torch.FloatTensor, alpha: float) -> torch.FloatTensor:
        raise NotImplementedError

class DeltaStrategy(WatermarkStrategy):
    """Strategy for delta-reweight."""
    def from_random(self, rng: Union[torch.Generator, list[torch.Generator]], vocab_size: int) -> torch.LongTensor:
        if isinstance(rng, list):
            batch_size = len(rng)
            u = torch.stack(
                [
                    torch.rand((), generator=rng[i], device=rng[i].device)
                    for i in range(batch_size)
                ]
            )
        else:
            u = torch.rand((), generator=rng, device=rng.device)
        return u
    
    def reweight_logits(self, u: torch.LongTensor, p_logits: torch.FloatTensor) -> torch.FloatTensor:
        """Reweight the logits using the u."""
        cumsum = torch.cumsum(F.softmax(p_logits, dim=-1), dim=-1)
        index = torch.searchsorted(cumsum, u[..., None], right=True)
        index = torch.clamp(index, 0, p_logits.shape[-1] - 1)
        modified_logits = torch.where(
            torch.arange(p_logits.shape[-1], device=p_logits.device) == index,
            torch.full_like(p_logits, 0),
            torch.full_like(p_logits, float("-inf")),
        )
        return modified_logits
    
class GammaStrategy(WatermarkStrategy):
    """Strategy for gamma-reweight."""
    def from_random(self, rng: Union[torch.Generator, list[torch.Generator]], vocab_size: int) -> torch.LongTensor:
        """Generate a permutation from the random number generator."""
        if isinstance(rng, list):
            batch_size = len(rng)
            shuffle = torch.stack(
                [
                    torch.randperm(vocab_size, generator=rng[i], device=rng[i].device)
                    for i in range(batch_size)
                ]
            )
        else:
            shuffle = torch.randperm(vocab_size, generator=rng, device=rng.device)
        return shuffle
        
    def reweight_logits(self, shuffle: torch.LongTensor, p_logits: torch.FloatTensor, alpha: float = 0.5) -> torch.FloatTensor:
        """Reweight the logits using the shuffle and alpha."""
        unshuffle = torch.argsort(shuffle, dim=-1)
        
        s_p_logits = torch.gather(p_logits, -1, shuffle)
        s_log_cumsum = torch.logcumsumexp(s_p_logits, dim=-1)
        
        # normalize the log_cumsum to force the last element to be 0
        s_log_cumsum = s_log_cumsum - s_log_cumsum[..., -1:]
        s_cumsum = torch.exp(s_log_cumsum)
        s_p = F.softmax(s_p_logits, dim=-1)

        boundary_1 = torch.argmax((s_cumsum > alpha).to(torch.int), dim=-1, keepdim=True)
        p_boundary_1 = torch.gather(s_p, -1, boundary_1)
        portion_in_right_1 = (torch.gather(s_cumsum, -1, boundary_1) - alpha) / p_boundary_1
        portion_in_right_1 = torch.clamp(portion_in_right_1, 0, 1)
        s_all_portion_in_right_1 = (s_cumsum > alpha).type_as(p_logits)
        s_all_portion_in_right_1.scatter_(-1, boundary_1, portion_in_right_1)

        boundary_2 = torch.argmax((s_cumsum > (1-alpha)).to(torch.int), dim=-1, keepdim=True)
        p_boundary_2 = torch.gather(s_p, -1, boundary_2)
        portion_in_right_2 = (torch.gather(s_cumsum, -1, boundary_2) - (1-alpha)) / p_boundary_2
        portion_in_right_2 = torch.clamp(portion_in_right_2, 0, 1)
        s_all_portion_in_right_2 = (s_cumsum > (1-alpha)).type_as(p_logits)
        s_all_portion_in_right_2.scatter_(-1, boundary_2, portion_in_right_2)

        s_all_portion_in_right = s_all_portion_in_right_2/2 + s_all_portion_in_right_1/2
        s_shift_logits = torch.log(s_all_portion_in_right)
        shift_logits = torch.gather(s_shift_logits, -1, unshuffle)

        return p_logits + shift_logits


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ================================================
# dip.py
# Description: Implementation of DiPmark algorithm
# ================================================

import torch
import hashlib
import random
from typing import Tuple, Union
import torch.nn.functional as F
from math import sqrt
from functools import partial
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization


class DIPConfig(BaseConfig):
    """Config class for DiP algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.gamma = self.config_dict['gamma']
        self.alpha = self.config_dict['alpha']
        self.ignore_history_generation = bool(self.config_dict['ignore_history_generation'])
        self.ignore_history_detection = bool(self.config_dict['ignore_history_detection'])
        self.z_threshold = self.config_dict['z_threshold']
        self.prefix_length = self.config_dict['prefix_length']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'DIP'

class DIPUtils:
    """Utility class for DiP algorithm, contains helper functions."""

    def __init__(self, config: DIPConfig, *args, **kwargs) -> None:
        """
            Initialize the DIP utility class.

            Parameters:
                config (DIPConfig): Configuration for the DiP algorithm.
        """
        self.config = config
        self.rng = torch.Generator(device=self.config.device)
        self.cc_history = set()
        self.state_indicator = 0 # 0 for generation, 1 for detection and visualization
        

    def _get_rng_seed(self, context_code: any) -> int:
        """Get the random seed from the given context code and private key."""
        if (
            (not self.config.ignore_history_generation and self.state_indicator == 0) or 
            (not self.config.ignore_history_detection and self.state_indicator == 1)
        ):
            self.cc_history.add(context_code)
            
        m = hashlib.sha256()
        m.update(context_code)
        m.update(self.config.hash_key)
        
        full_hash = m.digest()
        seed = int.from_bytes(full_hash, "big") % (2**32 - 1)
        return seed
    
    def _extract_context_code(self, context: torch.LongTensor) -> bytes:
        """Extract context code from the given context."""
        if self.config.prefix_length == 0:
            return context.detach().cpu().numpy().tobytes()
        else:
            return context[-self.config.prefix_length : ].detach().cpu().numpy().tobytes()
    
    def from_random(self, rng: Union[torch.Generator, list[torch.Generator]], vocab_size: int) -> torch.LongTensor:
        """Generate a permutation from the random number generator."""
        if isinstance(rng, list):
            batch_size = len(rng)
            shuffle = torch.stack(
                [
                    torch.randperm(vocab_size, generator=rng[i], device=rng[i].device)
                    for i in range(batch_size)
                ]
            )
        else:
            shuffle = torch.randperm(vocab_size, generator=rng, device=rng.device)
        return shuffle

    def reweight_logits(self, shuffle: torch.LongTensor, p_logits: torch.FloatTensor) -> torch.FloatTensor:
        """Reweight the logits using the shuffle and alpha."""
        unshuffle = torch.argsort(shuffle, dim=-1)
        
        s_p_logits = torch.gather(p_logits, -1, shuffle)
        s_log_cumsum = torch.logcumsumexp(s_p_logits, dim=-1)
        
        # normalize the log_cumsum to force the last element to be 0
        s_log_cumsum = s_log_cumsum - s_log_cumsum[..., -1:]
        s_cumsum = torch.exp(s_log_cumsum)
        s_p = F.softmax(s_p_logits, dim=-1)

        boundary_1 = torch.argmax((s_cumsum > self.config.alpha).to(torch.int), dim=-1, keepdim=True)
        p_boundary_1 = torch.gather(s_p, -1, boundary_1)
        portion_in_right_1 = (torch.gather(s_cumsum, -1, boundary_1) - self.config.alpha) / p_boundary_1
        portion_in_right_1 = torch.clamp(portion_in_right_1, 0, 1)
        s_all_portion_in_right_1 = (s_cumsum > self.config.alpha).type_as(p_logits)
        s_all_portion_in_right_1.scatter_(-1, boundary_1, portion_in_right_1)

        boundary_2 = torch.argmax((s_cumsum > (1-self.config.alpha)).to(torch.int), dim=-1, keepdim=True)
        p_boundary_2 = torch.gather(s_p, -1, boundary_2)
        portion_in_right_2 = (torch.gather(s_cumsum, -1, boundary_2) - (1-self.config.alpha)) / p_boundary_2
        portion_in_right_2 = torch.clamp(portion_in_right_2, 0, 1)
        s_all_portion_in_right_2 = (s_cumsum > (1-self.config.alpha)).type_as(p_logits)
        s_all_portion_in_right_2.scatter_(-1, boundary_2, portion_in_right_2)

        s_all_portion_in_right = s_all_portion_in_right_2/2 + s_all_portion_in_right_1/2
        s_shift_logits = torch.log(s_all_portion_in_right)
        shift_logits = torch.gather(s_shift_logits, -1, unshuffle)

        return p_logits + shift_logits
    
    def get_seed_for_cipher(self, input_ids: torch.LongTensor) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Get the mask and seeds for the cipher."""
        batch_size = input_ids.size(0)
        context_codes = [
            self._extract_context_code(input_ids[i]) for i in range(batch_size)
        ]

        mask, seeds = zip(
            *[
                (context_code in self.cc_history, self._get_rng_seed(context_code))
                for context_code in context_codes 
            ]
        )
        
        return mask, seeds
    
    def _get_green_token_quantile(self, input_ids: torch.LongTensor, vocab_size, current_token):
        """Get the vocab quantile of current token"""
        mask, seeds = self.get_seed_for_cipher(input_ids.unsqueeze(0))
        
        rng = [
            torch.Generator(device=input_ids.device).manual_seed(seed) for seed in seeds
        ]
        
        mask = torch.tensor(mask, device=input_ids.device)
        shuffle = self.from_random(
            rng, vocab_size
        )
        
        token_quantile = [(torch.where(shuffle[0] == current_token)[0] +1)/vocab_size]
        return token_quantile, mask
    
    def _get_dip_score(self, input_ids: torch.LongTensor, vocab_size):
        """Get the DiP score of the input_ids"""
        scores = torch.zeros(input_ids.shape, device=input_ids.device)
        
        for i in range(input_ids.shape[-1] - 1):
            pre = input_ids[ : i+1]
            cur = input_ids[i+1]
            token_quantile, mask = self._get_green_token_quantile(pre, vocab_size, cur)
            # if the current token is in the history and ignore_history_detection is False, set the score to -1
            if not self.config.ignore_history_detection and mask[0]: 
                scores[i + 1] = -1
            else:
                scores[i + 1] = torch.stack(token_quantile).reshape(-1)
        
        return scores
    
    def score_sequence(self, input_ids: torch.LongTensor) -> tuple[float, list[int]]:
        """Score the input_ids and return z_score and green_token_flags."""
        score = self._get_dip_score(input_ids, self.config.vocab_size)
        
        green_tokens = torch.sum(score >= self.config.gamma, dim=-1, keepdim=False)
        green_token_flags = torch.zeros_like(score)
        condition_indices = torch.nonzero(score >= self.config.gamma, as_tuple=False).reshape(-1)
        green_token_flags[condition_indices] = 1
        green_token_flags[:self.config.prefix_length] = -1
        
        # Use two different ways to calculate z_score depending on whether to ignore history
        if not self.config.ignore_history_detection:
            ignored_indices = torch.nonzero(score == -1, as_tuple=False).reshape(-1)
            
            # Visualize the ignored tokens as ignored
            green_token_flags[ignored_indices] = -1
            
            # Calculate z_score using the sequence length after ignoring the ignored tokens
            sequence_length_for_calculation = input_ids.size(-1) - ignored_indices.size(0)
            z_score = (green_tokens - (1-self.config.gamma) * sequence_length_for_calculation) / sqrt(sequence_length_for_calculation)
        else:
            z_score = (green_tokens - (1-self.config.gamma) * input_ids.size(-1)) / sqrt(input_ids.size(-1))
        
        return z_score.item(), green_token_flags.tolist()

class DIPLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for DiP algorithm, process logits to add watermark."""

    def __init__(self, config: DIPConfig, utils: DIPUtils, *args, **kwargs) -> None:
        """
            Initialize the DIP logits processor.

            Parameters:
                config (DIPConfig): Configuration for the DiP algorithm.
                utils (DIPUtils): Utility class for the DiP algorithm.
        """
        self.config = config
        self.utils = utils
    
    def _apply_watermark(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Apply watermark to the scores."""
        mask, seeds = self.utils.get_seed_for_cipher(input_ids)
        
        rng = [
            torch.Generator(device=scores.device).manual_seed(seed) for seed in seeds
        ]
        mask = torch.tensor(mask, device=scores.device)
        shuffle = self.utils.from_random(
            rng, scores.size(1)
        )

        reweighted_scores = self.utils.reweight_logits(shuffle, scores)
        
        return mask, reweighted_scores

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process logits to add watermark."""
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores
        
        mask, reweighted_scores = self._apply_watermark(input_ids, scores)
        
        if self.config.ignore_history_generation:
            return reweighted_scores
        else:
            return torch.where(mask[:, None], scores, reweighted_scores)
    

class DIP(BaseWatermark):
    """Top-level class for DIP algorithm."""

    def __init__(self, algorithm_config: str | DIPConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the DIP algorithm.

            Parameters:
                algorithm_config (str | DIPConfig): Path to the algorithm configuration file or DIPConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = DIPConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, DIPConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a DIPConfig instance")
        
        self.utils = DIPUtils(self.config)
        self.logits_processor = DIPLogitsProcessor(self.config, self.utils)
    
    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text."""

        # Set the state indicator to 0 for generation
        self.utils.state_indicator = 0

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # Generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # Decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        # Clear the history
        self.utils.cc_history.clear()
        return watermarked_text
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""
        
        # Set the state indicator to 1 for detection
        self.utils.state_indicator = 1

        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)

        # Compute z-score using a utility method
        z_score, _ = self.utils.score_sequence(encoded_text)
        
        # Determine if the z-score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Clear the history
        self.utils.cc_history.clear()

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)
        
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
        """Get data for visualization."""
        
        # Set the state indicator to 1 for visualization
        self.utils.state_indicator = 1
        
        # Encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
        # Compute the highlight values
        _, highlight_values = self.utils.score_sequence(encoded_text)
        
        # decode single tokens
        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)
        
        # Clear the history
        self.utils.cc_history.clear()
        
        return DataForVisualization(decoded_tokens, highlight_values)




# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ==============================================
# xsir.py
# Description: Implementation of X-SIR algorithm
# ==============================================

import os
import json
import torch
import numpy as np
from typing import Union
from functools import partial
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from .transform_model import TransformModel
from sentence_transformers import SentenceTransformer
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization
from .generate_semantic_mappings import generate_semantic_mappings


class XSIRConfig(BaseConfig):
    """Config class for X-SIR algorithm, load config file and initialize parameters."""

    # def __init__(self, algorithm_config: str, transformers_config: TransformersConfig, *args, **kwargs) -> None:
    #     """
    #         Initialize the X-SIR configuration.

    #         Parameters:
    #             algorithm_config (str): Path to the algorithm configuration file.
    #             transformers_config (TransformersConfig): Configuration for the transformers model.
    #     """
    #     if algorithm_config is None:
    #         config_dict = load_config_file('config/XSIR.json')
    #     else:
    #         config_dict = load_config_file(algorithm_config)
    #     if config_dict['algorithm_name'] != 'XSIR':
    #         raise AlgorithmNameMismatchError('XSIR', config_dict['algorithm_name'])

    #     # Load config
    #     self.delta = config_dict['delta']
    #     self.chunk_length = config_dict['chunk_length']
    #     self.scale_dimension = config_dict['scale_dimension']
    #     self.z_threshold = config_dict['z_threshold']
    #     self.transform_model_input_dim = config_dict['transform_model_input_dim']
    #     self.transform_model_name = config_dict['transform_model_name']
    #     self.embedding_model_path = config_dict['embedding_model_path']
    #     self.mapping_name = config_dict['mapping_name']
    #     self.dictionary = config_dict.get('dictionary', None)

    #     # Load transformer model's configuration
    #     self.generation_model = transformers_config.model
    #     self.generation_tokenizer = transformers_config.tokenizer
    #     self.vocab_size = transformers_config.vocab_size
    #     self.device = transformers_config.device
    #     self.gen_kwargs = transformers_config.gen_kwargs
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.delta = self.config_dict['delta']
        self.chunk_length = self.config_dict['chunk_length']
        self.scale_dimension = self.config_dict['scale_dimension']
        self.z_threshold = self.config_dict['z_threshold']
        self.transform_model_input_dim = self.config_dict['transform_model_input_dim']
        self.transform_model_name = self.config_dict['transform_model_name']
        self.embedding_model_path = self.config_dict['embedding_model_path']
        self.mapping_name = self.config_dict['mapping_name']
        self.dictionary = self.config_dict.get('dictionary', None)  
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'XSIR'


class XSIRUtils:
    """Utility class for X-SIR algorithm, contains helper functions."""

    def __init__(self, config: XSIRConfig, *args, **kwargs) -> None:
        """
            Initialize the X-SIR utility class.

            Parameters:
                config (XSIRConfig): Configuration for the SIR algorithm.
        """
        self.config = config
        # import ipdb; ipdb.set_trace()
        # Load the embedding model: mapping a sentence to an embedding
        self.embedding_model = SentenceTransformer(self.config.embedding_model_path).to(self.config.device)

        # Load the transform model: mapping a sentence embedding to logit bias
        self.transform_model = self._get_transform_model(
            model_name=self.config.transform_model_name,
            input_dim=config.transform_model_input_dim
        ).to(self.config.device, dtype=torch.float32)
        self.mapping = self._get_mapping(self.config.mapping_name)

    def get_embedding(self, sentence: str) -> torch.FloatTensor:
        """Get the embedding of the input sentence."""
        emb = self.embedding_model.encode(sentence, show_progress_bar=False, convert_to_tensor=True).to(self.config.device)
        return emb[None, :]

    def get_text_split(self, sentence: str) -> list[list[str]]:
        """Get the input sentence split into chunks."""
        tokens = self.config.generation_tokenizer.tokenize(sentence, add_special_tokens=False)
        return [tokens[x: x + self.config.chunk_length] for x in range(0, len(tokens), self.config.chunk_length)]

    def scale_vector(self, v: np.array) -> np.array:
        """Scale the input vector."""
        mean = np.mean(v)
        v_minus_mean = v - mean
        v_minus_mean = np.tanh(1000 * v_minus_mean)
        return v_minus_mean

    def get_context_sentence(self, input_ids: torch.LongTensor) -> str:
        """Get the context sentence from the input_ids."""
        input_sentence = self.config.generation_tokenizer.decode(input_ids, skip_special_tokens=True)
        input_tokens = self.config.generation_tokenizer.tokenize(input_sentence, add_special_tokens=False)

        # Split the input sentence into chunks
        chunks = [input_tokens[x: x + self.config.chunk_length] for x in range(0, len(input_tokens), self.config.chunk_length)]

        if len(chunks) == 0:
            return ""
        
        # If the last chunk is of the same length as the chunk length, return the whole sentence
        if len(chunks[-1]) == self.config.chunk_length:
            return input_sentence
        else:
            # Otherwise, return the sentence without the last chunk
            return self.config.generation_tokenizer.convert_tokens_to_string([tok for group in chunks[:-1] for tok in group])

    def _get_mapping(self, mapping_name: str) -> list[int]:
        """Get the mapping."""

        # try loading mapping from the provided mapping path
        try:
            with open(mapping_name, 'r') as f:
                mapping = json.load(f)

        # if the file does not exist, create a new mapping and save it to the provided mapping path
        except Exception as e:
            print(f"Error loading provided mapping file ({mapping_name}): {e}")
            print(f"Try to generate new mapping file ...")

            assert self.config.dictionary is not None and os.path.isfile(self.config.dictionary), \
                "Please provide a valid dictionary file for generating the mapping."
            assert self.config.scale_dimension is not None and self.config.scale_dimension > 0, \
                "Please provide a valid scale dimension for generating the mapping."

            mapping = generate_semantic_mappings(
                tokenizer=self.config.generation_tokenizer,
                dictionary=self.config.dictionary,
                output_file=mapping_name,
                dimension=self.config.scale_dimension,
                vocab_size=self.config.vocab_size
            )

            print(f"Generated new mapping file: {mapping_name}")
        return mapping

    def _get_transform_model(self, model_name: str, input_dim: int) -> TransformModel:
        """Get the transform model from the provided model name."""
        model = TransformModel(input_dim=input_dim)
        model.load_state_dict(torch.load(model_name))
        return model


class XSIRLogitsProcessor(LogitsProcessor):
    """Logits processor for X-SIR algorithm."""

    def __init__(self, config: XSIRConfig, utils: XSIRUtils, *args, **kwargs):
        """
            Initialize the X-SIR logits processor.

            Parameters:
                config (XSIRConfig): Configuration for the X-SIR algorithm.
                utils (XSIRUtils): Utility class for the X-SIR algorithm.
        """
        self.config = config
        self.utils = utils

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process logits to add watermark."""
        batched_bias = [None for _ in range(input_ids.shape[0])]

        for b_idx in range(input_ids.shape[0]):
            current_bias = self._get_bias(input_ids[b_idx])
            batched_bias[b_idx] = current_bias

        scores = self._bias_logits(scores=scores, batched_bias=batched_bias)
        return scores

    def _bias_logits(self, scores: torch.Tensor, batched_bias: list[np.ndarray]) -> torch.Tensor:
        """Bias the scores for the greenlist tokens."""
        batched_bias_np = np.array(batched_bias) 
        batched_bias_tensor = torch.Tensor(batched_bias_np).to(self.config.device)
        scores = scores + batched_bias_tensor * self.config.delta
        return scores

    def _get_bias(self, input_ids: torch.LongTensor) -> np.ndarray:
        """Calculate the bias for the given input_ids."""
        context_sentence = self.utils.get_context_sentence(input_ids)
        context_embedding = self.utils.get_embedding(context_sentence)
        output = self.utils.transform_model(context_embedding).cpu()[0].numpy()
        similarity_array = self.utils.scale_vector(output)[self.utils.mapping]
        return -similarity_array


class XSIR(BaseWatermark):
    """Top-level class for X-SIR algorithm."""

    def __init__(self, algorithm_config: str | XSIRConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the X-SIR algorithm.

            Parameters:
                algorithm_config (str | XSIRConfig): Path to the algorithm configuration file or XSIRConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = XSIRConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, XSIRConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a XSIRConfig instance")
        
        self.utils = XSIRUtils(self.config)
        self.logits_processor = XSIRLogitsProcessor(self.config, self.utils)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs):
        """Generate watermarked text."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )

        # encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)

        # generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)

        # decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]

        return watermarked_text

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs) -> Union[dict[str, Union[bool, float]], tuple[bool, float]]:
        """Detect watermark in the input text."""
        z_score = None
        is_watermarked = None

        all_value = [] # store all bias values
        chunks = self.utils.get_text_split(text) # split text into chunks

        if len(chunks) <= 1:
            print("Text is too short to detect watermark. Returning False.")
            is_watermarked, z_score = False, 0

        # calculate bias for each token
        for cid in range(1, len(chunks)):
            context_sentence = self.config.generation_tokenizer.convert_tokens_to_string([t for c in chunks[0: cid] for t in c])
            context_embedding = self.utils.get_embedding(context_sentence)
            output = self.utils.transform_model(context_embedding).cpu().detach()[0].numpy()
            similarity_array = self.utils.scale_vector(output)[self.utils.mapping]

            tokens = chunks[cid]
            token_ids = self.config.generation_tokenizer.convert_tokens_to_ids(tokens)

            for tok_ids in token_ids:
                all_value.append(-float(similarity_array[tok_ids]))

        # z_score is the mean of all bias values
        z_score = np.mean(all_value)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold # TODO: what if we have no `z_threshold`?

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)

    def get_data_for_visualization(self, text: str, *args, **kwargs):
        """Get data for visualization."""
        highlight_values = []
        decoded_tokens = []
        chunks = self.utils.get_text_split(text) # split text into chunks

        if len(chunks) <= 1:
            # If the sentence is shorter than required, append highlight -1 for each token
            decoded_tokens = chunks[0]
            highlight_values = [-1] * len(decoded_tokens)

        # calculate bias for each token
        for cid in range(1, len(chunks)):
            context_sentence = self.config.generation_tokenizer.convert_tokens_to_string([t for c in chunks[0: cid] for t in c])
            context_embedding = self.utils.get_embedding(context_sentence)
            output = self.utils.transform_model(context_embedding).cpu().detach()[0].numpy()
            similarity_array = self.utils.scale_vector(output)[self.utils.mapping]

            tokens = chunks[cid]
            token_ids = self.config.generation_tokenizer.convert_tokens_to_ids(tokens)

            for tok_ids in token_ids:
                similarity_value = -float(similarity_array[tok_ids])
                highlight_values.append(1 if similarity_value > 0 else 0)
                decoded_tokens.append(self.config.generation_tokenizer.convert_ids_to_tokens(tok_ids))

        return DataForVisualization(decoded_tokens, highlight_values)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# exp.py
# Description: Implementation of EXP algorithm
# ============================================

import torch
import scipy
from math import log
from ..base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from visualize.data_for_visualization import DataForVisualization


class EXPConfig(BaseConfig):
    """Config class for EXP algorithm, load config file and initialize parameters."""

    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.prefix_length = self.config_dict['prefix_length']
        self.hash_key = self.config_dict['hash_key']
        self.threshold = self.config_dict['threshold']
        self.sequence_length = self.config_dict['sequence_length']
        self.top_k = getattr(self.transformers_config, 'top_k', -1)
        self.temperature = getattr(self.transformers_config, 'temperature', 0.7)
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'EXP'


class EXPUtils:
    """Utility class for EXP algorithm, contains helper functions."""

    def __init__(self, config: EXPConfig, *args, **kwargs) -> None:
        """
            Initialize the EXP utility class.

            Parameters:
                config (EXPConfig): Configuration for the EXP algorithm.
        """
        self.config = config
        self.rng = torch.Generator()

    def seed_rng(self, input_ids: torch.LongTensor) -> None:
        """Seed the random number generator with the last `prefix_length` tokens of the input."""
        time_result = 1
        for i in range(0, self.config.prefix_length):
            time_result *= input_ids[-1 - i].item()
        prev_token = time_result % self.config.vocab_size
        self.rng.manual_seed(self.config.hash_key * prev_token)
        return
    
    def exp_sampling(self, probs: torch.Tensor, u: torch.Tensor) -> torch.Tensor:
        """Sample a token from the vocabulary using the exponential sampling method."""
        
        # If top_k is not specified, use argmax
        if self.config.top_k <= 0:
            return torch.argmax(u ** (1 / probs), axis=1).unsqueeze(-1)
        
        # Ensure top_k is not greater than the vocabulary size
        top_k = min(self.config.top_k, probs.size(-1))
    
        # Get the top_k probabilities and their indices
        top_probs, top_indices = torch.topk(probs, top_k, dim=-1)
    
        # Perform exponential sampling on the top_k probabilities
        sampled_indices = torch.argmax(u.gather(-1, top_indices) ** (1 / top_probs), dim=-1)
    
        # Map back the sampled indices to the original vocabulary indices
        return top_indices.gather(-1, sampled_indices.unsqueeze(-1))
    
    def _value_transformation(self, value):
        """Transform the value to a range between 0 and 1."""
        return value/(value + 1)
    

class EXP(BaseWatermark):
    """Top-level class for the EXP algorithm."""

    def __init__(self, algorithm_config: str | EXPConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the EXP algorithm.

            Parameters:
                algorithm_config (str | EXPConfig): Path to the algorithm configuration file or EXPConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = EXPConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, EXPConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a EXPConfig instance")
        
        self.utils = EXPUtils(self.config)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text using the EXP algorithm."""

        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer.encode(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        
        # Initialize
        inputs = encoded_prompt
        attn = torch.ones_like(encoded_prompt)
        past = None

        # Generate tokens
        for i in range(self.config.sequence_length):
            with torch.no_grad():
                if past:
                    output = self.config.generation_model(inputs[:,-1:], past_key_values=past, attention_mask=attn)
                else:
                    output = self.config.generation_model(inputs)
            
            # Get probabilities with temperature
            probs = torch.nn.functional.softmax(output.logits[:,-1, :self.config.vocab_size] / self.config.temperature, dim=-1).cpu()
            
            # Generate r1, r2,..., rk
            self.utils.seed_rng(inputs[0])
            random_numbers = torch.rand(self.config.vocab_size, generator=self.utils.rng)
            
            # Sample token to add watermark
            token = self.utils.exp_sampling(probs, random_numbers).to(self.config.device)

            # Update inputs
            inputs = torch.cat([inputs, token], dim=-1)

            # Update past
            past = output.past_key_values

            # Update attention mask
            attn = torch.cat([attn, attn.new_ones((attn.shape[0], 1))], dim=-1)
        
        watermarked_tokens = inputs[0].detach().cpu()
        watermarked_text = self.config.generation_tokenizer.decode(watermarked_tokens, skip_special_tokens=True)

        return watermarked_text    

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs) -> dict:
        """Detect watermark in the text."""

        # Encode the text into tokens using the configured tokenizer
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False).numpy()[0]

        # Calculate the number of tokens to score, excluding the prefix
        num_scored = len(encoded_text) - self.config.prefix_length
        total_score = 0

        for i in range(self.config.prefix_length, len(encoded_text)):
            # Seed RNG with the prefix of the encoded text
            self.utils.seed_rng(encoded_text[:i])

            # Generate random numbers for each token in the vocabulary
            random_numbers = torch.rand(self.config.vocab_size, generator=self.utils.rng)

            # Calculate score for the current token
            r = random_numbers[encoded_text[i]]
            total_score += log(1 / (1 - r))

        # Calculate p_value
        p_value = scipy.stats.gamma.sf(total_score, num_scored, loc=0, scale=1)

        # Determine if the computed score exceeds the threshold for watermarking
        is_watermarked = p_value < self.config.threshold

        # Return results based on the `return_dict` flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": p_value}
        else:
            return (is_watermarked, p_value)
        
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> DataForVisualization:
        """Get data for visualization."""

        # Encode the text into tokens using the configured tokenizer
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False).numpy()[0]

        # Initialize the list of values with None for the prefix length
        highlight_values = [None] * self.config.prefix_length

        # Calculate the value for each token beyond the prefix
        for i in range(self.config.prefix_length, len(encoded_text)):
            # Seed the random number generator using the prefix of the encoded text
            self.utils.seed_rng(encoded_text[:i])
            random_numbers = torch.rand(self.config.vocab_size, generator=self.utils.rng)
            r = random_numbers[encoded_text[i]]
            v = log(1 / (1 - r))
            v = self.utils._value_transformation(v)
            highlight_values.append(v)

        # Decode each token id to its corresponding string token
        decoded_tokens = [self.config.generation_tokenizer.decode([token_id]) for token_id in encoded_text]
        
        return DataForVisualization(decoded_tokens, highlight_values)
        
    


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import random
import argparse
import networkx as nx
from transformers import AutoTokenizer, PreTrainedTokenizer

# Maximum size of connected components
# Note: The returned connected components are not guaranteed to be smaller than the max_size constraint.
#       We only use max_size to determine the number of clusters.
CC_MAX_SIZE=250

def parse_args():
    parser = argparse.ArgumentParser(description='Generate mappings.')
    parser.add_argument('--model', type=str, required=True, help='Model name')
    parser.add_argument('--dictionary', type=str, required=True, help='Dictionary path. One line per entry.')
    parser.add_argument('--output_file', type=str, required=True, help='Output file path')
    parser.add_argument('--dimension', type=int, default=300, help='Dimension of the embeddings')
    return parser.parse_args()

def generate_semantic_mappings(
    dictionary: str,
    output_file: str,
    dimension: int = 300,
    model: str = None,
    tokenizer: PreTrainedTokenizer = None,
    vocab_size: int = None
):
    # set seed
    random.seed(0)

    # Load tokenizer
    assert model is not None or tokenizer is not None, "One of model or tokenizer should be provided"
    assert model is None or tokenizer is None, "Only one of model or tokenizer should be provided"

    if model is not None:
        tokenizer = AutoTokenizer.from_pretrained(model, trust_remote_code=True)

    if vocab_size is not None and vocab_size > tokenizer.vocab_size:
        print(f"Warning: vocab_size={vocab_size} is larger than tokenizer.vocab_size={tokenizer.vocab_size}")
        print(f"Adding {vocab_size - tokenizer.vocab_size} extra tokens")
        tokenizer.add_tokens([f"<extra_id_{i}>" for i in range(vocab_size - tokenizer.vocab_size)])

    vocab = tokenizer.get_vocab()

    # Load dictionary as edges
    edges = []
    with open(dictionary) as f:
        for line in f:
            src_token, tgt_token = line.split()
            edges.append((src_token, tgt_token))

    # Add self-loop for each token
    for token in vocab:
        edges.append((token, token))

    # Build graph & find connected components
    graph = nx.Graph(edges)
    connected_components_node = list(nx.connected_components(graph)) # list of nodes
    connected_components_node.sort(key=lambda x: len(x), reverse=True) # sort by size
    connected_components_graph = [graph.subgraph(ccn) for ccn in connected_components_node] # list of graphs

    # Split connected components into clusters
    clusters = []
    for ccg in connected_components_graph:
        if len(ccg) <= CC_MAX_SIZE:
            clusters.append(list(ccg))
        else:
            resolution = 1
            if len(ccg) > 10000: # use higher resolution for extremely large connected components to find smaller clusters
                resolution = 10
            cs = nx.community.louvain_communities(ccg, seed=0, resolution=resolution)
            clusters.extend(cs)
            print(f"Splitted {len(ccg)} nodes into {len(cs)} clusters")

    # Valid clusters: filter tokens that are not in the vocabulary
    valid_clusters = []
    for c in clusters:
        valid_c = set()
        for token in c:
            if token in vocab:
                valid_c.add(token)
        if len(valid_c) > 0:
            valid_clusters.append(list(valid_c))

    # Check if all tokens are included
    all_valid_tokens = []
    for c in valid_clusters:
        all_valid_tokens.extend(c)
    assert len(all_valid_tokens) == len(vocab), f"len(all_valid_tokens)={len(all_valid_tokens)} len(vocab)={len(vocab)}"

    # Generate mappings
    clustre_mapping = [random.randint(0, dimension - 1) for _ in range(len(valid_clusters))]
    mapping = [None for _ in range(len(vocab))]
    for i, c in enumerate(valid_clusters):
        for token in c:
            token_id = tokenizer.convert_tokens_to_ids(token)
            assert isinstance(token_id, int)
            assert mapping[token_id] is None
            mapping[token_id] = clustre_mapping[i]
    assert all(x is not None for x in mapping)

    # Save mappings
    if os.path.dirname(output_file) != '' and (not os.path.exists(os.path.dirname(output_file))):
        os.makedirs(os.path.dirname(output_file))
    with open(output_file, 'w') as f:
        json.dump(mapping, f, indent=4)

    # Save clusters
    output_cluster_file = ".".join(output_file.split(".")[:-1]) + "_clusters.json"
    with open(output_cluster_file, "w") as f:
        # sort by size
        valid_clusters.sort(key=lambda x: len(x), reverse=True)
        json.dump(valid_clusters, f, indent=4, ensure_ascii=False)

    # Print statistics
    print(f"Vocabulary size: {len(vocab)}")
    print(f"Number of clusters: {len(valid_clusters)}")
    print(f"Number of clusters with size >= 2: {len([cc for cc in valid_clusters if len(cc) >= 2])}")
    print(f"Number of tokens in the cc with size >= 2: {sum([len(cc) for cc in valid_clusters if len(cc) >= 2])}")
    print(f"Top 20 largest cc: {[len(cc) for cc in sorted(connected_components_graph, key=lambda x: len(x), reverse=True)[:20]]}")
    print(f"Top 20 largest clusters: {[len(cc) for cc in sorted(clusters, key=lambda x: len(x), reverse=True)[:20]]}")
    print(f"Top 20 largest valid clusters: {[len(cc) for cc in sorted(valid_clusters, key=lambda x: len(x), reverse=True)[:20]]}")
    print(f"Vocab coverage: {sum([len(cc) for cc in valid_clusters if len(cc) >= 2]) / len(vocab) * 100:.2f}%")

    return mapping

if __name__ == '__main__':
    args = parse_args()
    generate_semantic_mappings(
        model=args.model,
        dictionary=args.dictionary,
        output_file=args.output_file,
        dimension=args.dimension
    )


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ================================================
# exp_edit.py
# Description: Implementation of EXPEdit algorithm
# ================================================

import torch
import numpy as np
from math import log
from ..base import BaseWatermark, BaseConfig
from .mersenne import MersenneRNG
from utils.utils import load_config_file
from .cython_files.levenshtein import levenshtein
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from visualize.data_for_visualization import DataForVisualization


class ITSEditConfig(BaseConfig):
    """Config class for EXPEdit algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.pseudo_length = self.config_dict['pseudo_length']
        self.sequence_length = self.config_dict['sequence_length']
        self.n_runs = self.config_dict['n_runs']
        self.p_threshold = self.config_dict['p_threshold']
        self.key = self.config_dict['key']
        self.top_k = self.config_dict['top_k']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'ITSEdit'


class ITSEditUtils:
    """Utility class for EXPEdit algorithm, contains helper functions."""

    def __init__(self, config: ITSEditConfig, *args, **kwargs) -> None:
        """
            Initialize the EXPEdit utility class.

            Parameters:
                config (EXPEditConfig): Configuration for the EXPEdit algorithm.
        """
        self.config = config
        torch.manual_seed(self.config.key)
        self.rng = MersenneRNG(self.config.key)

    def transform_sampling(self,probs,pi,xi):
        """Sample token using inverse transform method."""
        cdf = torch.cumsum(torch.gather(probs, 1, pi), 1)
        return torch.gather(pi, 1, torch.searchsorted(cdf, xi))
    
    def transform_key_func(self,generator,n,vocab_size):
        """Generate key for watermark."""
        pi = torch.randperm(vocab_size, generator=generator)
        xi = torch.rand((n,1), generator=generator)
        return xi,pi
    
    def value_transformation(self, value: float) -> float:
        """Transform value to range [0, 1]."""
        return value/(value + 1)

    def phi(self,tokens,n,k,generator,key_func,vocab_size,dist,null=False,normalize=False):
        '''Compute the test statistic for the watermark detection algorithm.'''
        eff_vocab_size = vocab_size
        xi,pi = key_func(generator,n,vocab_size)
        tokens = torch.argsort(pi)[tokens]
        if normalize:
            tokens = tokens.float() / vocab_size
        
        A = self.adjacency(tokens,xi,dist,k)
        closest = torch.min(A,axis=1)[0]
        min_value, min_index = torch.min(closest, 0)
        return min_value, min_index.item()

    def adjacency(self,tokens,xi,dist,k):
        '''Compute the adjacency matrix for the test statistic.'''
        m = len(tokens)
        n = len(xi)
        A = torch.empty(size=(m-(k-1),n))
        for i in range(m-(k-1)):
            for j in range(n):
                A[i][j] = dist(tokens[i:i+k],xi[(j+torch.arange(k))%n])

        return A

class ITSEdit(BaseWatermark):
    """Top-level class for the EXPEdit algorithm."""
    def __init__(self, algorithm_config: str, transformers_config: TransformersConfig, *args, **kwargs) -> None:
        self.config = ITSEditConfig(algorithm_config, transformers_config)
        self.utils = ITSEditUtils(self.config)
    
    def generate(self,model,prompts,vocab_size,n,m,seeds,key_func,sampler,random_offset=False):
        '''Generate watermark tokens.'''
        batch_size = len(prompts)
        generator = torch.Generator()
        xis,pis = [],[]
        # Generate keys
        for seed in seeds:
            generator.manual_seed(int(seed))
            xi,pi = key_func(generator,n,vocab_size)
            xis.append(xi.unsqueeze(0))
            pis.append(pi.unsqueeze(0))
        xis = torch.vstack(xis)
        pis = torch.vstack(pis)
        offset = torch.zeros(size=(batch_size,),dtype=torch.int64)
        inputs = prompts.to(model.device)
        attn = torch.ones_like(inputs)
        past = None
        # Generate tokens
        for i in range(m):
            with torch.no_grad():
                if past:
                    output = model(inputs[:,-1:], past_key_values=past, attention_mask=attn)
                else:
                    output = model(inputs)
            probs = torch.nn.functional.softmax(output.logits[:,-1], dim=-1).cpu()
            tokens = sampler(probs, pis, xis[torch.arange(batch_size),(offset.squeeze()+i)%n]).to(model.device)
            inputs = torch.cat([inputs, tokens], dim=-1)

            past = output.past_key_values
            attn = torch.cat([attn, attn.new_ones((attn.shape[0], 1))], dim=-1)

        return inputs.detach().cpu()
    
    def clip(self,tokens):
        '''Clip tokens to remove padding.'''
        eos = torch.where(tokens == 2)[0] # find instances of the EOS token
        if len(eos) > 0:
            # truncate after the first EOS token (end of the response)
            tokens = tokens[:eos[0]]

        return tokens
    
    def generate_unwatermarked_text(self, prompt: str, *args, **kwargs) -> str:
        '''Generate unwatermarked text.'''
        def generate_rnd(prompts,m,model):
            inputs = prompts.to(model.device)
            attn = torch.ones_like(inputs)
            past = None
            for i in range(m):
                torch.manual_seed(self.config.key)
                with torch.no_grad():
                    output = model(inputs)

                probs = torch.nn.functional.softmax(output.logits[:,-1], dim=-1)
                tokens = torch.multinomial(probs,1)
                inputs = torch.cat([inputs, tokens], dim=1)

                past = output.past_key_values
                attn = torch.cat([attn, attn.new_ones((attn.shape[0], 1))], dim=-1)
            
            return inputs.detach().cpu()
        encoded_prompt = self.config.generation_tokenizer.encode(prompt, return_tensors='pt', truncation=True, max_length=2048)[0].to(self.config.device)
        prompt_tokens = len(encoded_prompt)
        unwatermarked_prompt=generate_rnd(encoded_prompt.unsqueeze(0),self.config.sequence_length,self.config.generation_model)[0,prompt_tokens:]
        return self.config.generation_tokenizer.decode(unwatermarked_prompt, skip_special_tokens=True)
    
    def generate_watermarked_text(self, prompt: str, *args, **kwargs):
        """Generate watermarked text."""
        seeds = torch.randint(2**32, (self.config.sequence_length,))
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer.encode(prompt, return_tensors='pt', truncation=True, max_length=2048)[0].to(self.config.device)
        prompt_tokens = len(encoded_prompt)
        generate_watermark = lambda prompt,seed : self.generate(self.config.generation_model,
                                                       prompt,
                                                       self.config.vocab_size,
                                                       self.config.pseudo_length,
                                                       self.config.sequence_length,
                                                       seed,
                                                       self.utils.transform_key_func,
                                                       self.utils.transform_sampling,
                                                       random_offset=True)
        watermark_sample=generate_watermark(encoded_prompt.unsqueeze(0), seeds[0].unsqueeze(0))[0,prompt_tokens:]
        watermarked_sample = self.clip(watermark_sample)
        return self.config.generation_tokenizer.decode(watermarked_sample, skip_special_tokens=True)
    
    def permutation_test(self,tokens,vocab_size,n,k,seed,test_stat,n_runs=100,max_seed=100000):
        '''Permutation test for watermark detection.'''
        torch.manual_seed(self.config.key)
        generator = torch.Generator()
        generator.manual_seed(int(seed))

        test_result,_= test_stat(tokens=tokens,
                                n=n,
                                k=k,
                                generator=generator,
                                vocab_size=vocab_size)
        p_val = 0
        for run in range(n_runs):
            pi = torch.randperm(vocab_size)
            tokens = torch.argsort(pi)[tokens]
            
            seed = torch.randint(high=max_seed,size=(1,)).item()
            generator.manual_seed(int(seed))
            null_result,_ = test_stat(tokens=tokens,
                                    n=n,
                                    k=k,
                                    generator=generator,
                                    vocab_size=vocab_size,
                                    null=True)
            # assuming lower test values indicate presence of watermark
            p_val += (null_result <= test_result).float() / n_runs
        
        return p_val
    
    def transform_edit_score(self,tokens,xi,gamma=1):
        '''Compute the edit score between two sequences.'''
        return levenshtein(tokens.numpy(),xi.squeeze().numpy(),gamma)
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""
        torch.manual_seed(self.config.key)
        seeds = torch.randint(2**32, (self.config.sequence_length,))
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', truncation=True,max_length=2048)[0]
        test_stat = lambda tokens,n,k,generator,vocab_size,null=False : self.utils.phi(tokens=tokens,
                                                                        n=n,
                                                                        k=k,
                                                                        generator=generator,
                                                                        key_func=self.utils.transform_key_func,
                                                                        vocab_size=self.config.vocab_size,
                                                                        dist=lambda x,y : self.transform_edit_score(x,y,1),
                                                                        null=False,
                                                                        normalize=True)
        test = lambda tokens,seed : self.permutation_test(tokens,
                                             self.config.vocab_size,
                                             self.config.pseudo_length,
                                             len(tokens),
                                             seed,
                                             test_stat,
                                             n_runs=self.config.n_runs,)
        pval_watermark = test(encoded_text, seeds[0])
        if return_dict:
            return {"is_watermarked": (pval_watermark<=self.config.p_threshold).item(), "score": pval_watermark.item()}
        else:
            return ((pval_watermark<=self.config.p_threshold).item(), pval_watermark.item())
    
    def get_data_for_visualization(self, text: str, *args, **kwargs):
        """Get data for visualization."""

        pass


import torch
import torch.nn as nn
import numpy as np


class DeltaNetwork(nn.Module):
    def __init__(self, input_dim=2048, layers=2, init_val=2.0):
        super(DeltaNetwork, self).__init__()
        if layers == 2:
            self.delta = nn.Sequential(
                nn.Linear(input_dim, 64),
                nn.LeakyReLU(),
                nn.Linear(64, 1)
            )
        elif layers == 3:
            self.delta = nn.Sequential(
                nn.Linear(input_dim, 256),
                nn.LeakyReLU(),
                nn.Linear(256, 32),
                nn.LeakyReLU(),
                nn.Linear(32, 1)
            )
        elif layers == 5:
            self.delta = nn.Sequential(
                nn.Linear(input_dim, 512),
                nn.LeakyReLU(),
                nn.Linear(512, 128),
                nn.LeakyReLU(),
                nn.Linear(128, 64),
                nn.LeakyReLU(),
                nn.Linear(64, 16),
                nn.LeakyReLU(),
                nn.Linear(16, 1)
            )

        for layer in self.delta:
            if isinstance(layer, nn.Linear):
                torch.nn.init.kaiming_normal_(layer.weight)
                torch.nn.init.zeros_(layer.bias)
        self.init_val = init_val
        nn.init.constant_(self.delta[-1].bias, init_val)  # Set bias to the calculated value
    def forward(self, x):
        return self.delta(x)

class GammaNetwork(nn.Module):
    def __init__(self, input_dim=2048, layers=2, init_val=0.25):
        super(GammaNetwork, self).__init__()
        if layers == 2:
            self.gamma = nn.Sequential(
                nn.Linear(input_dim, 64),
                nn.LeakyReLU(),
                nn.Linear(64, 1),
                nn.Sigmoid()
            )
        elif layers == 3:
            self.gamma = nn.Sequential(
                nn.Linear(input_dim, 256),
                nn.LeakyReLU(),
                nn.Linear(256, 32),
                nn.LeakyReLU(),
                nn.Linear(32, 1),
                nn.Sigmoid()
            )
        elif layers == 5:
            self.gamma = nn.Sequential(
                nn.Linear(input_dim, 512),
                nn.LeakyReLU(),
                nn.Linear(512, 128),
                nn.LeakyReLU(),
                nn.Linear(128, 64),
                nn.LeakyReLU(),
                nn.Linear(64, 16),
                nn.LeakyReLU(),
                nn.Linear(16, 1),
                nn.Sigmoid()
            )

        for layer in self.gamma:
            if isinstance(layer, nn.Linear):
                torch.nn.init.kaiming_normal_(layer.weight)
                torch.nn.init.zeros_(layer.bias)
        self.init_val = init_val
        nn.init.constant_(self.gamma[-2].bias, np.log(init_val / (1 - init_val)))  # Set bias to the calculated value
    def forward(self, x):
        return self.gamma(x)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# detector.py
# Description: Implementation of SynthID watermark detectors
# ============================================

import os
import abc
import tqdm
import json
import torch
import numpy as np
from .detector_bayesian_torch import RawBayesianDetector
from evaluation.dataset import C4Dataset


class SynthIDDetector(abc.ABC):
    """Base class for SynthID watermark detectors.
    
    This class defines the interface that all SynthID watermark detectors must implement.
    Subclasses should override the detect() method to implement specific detection algorithms.
    """

    def __init__(self):
        """Initialize the detector."""
        pass

    @abc.abstractmethod
    def detect(self, g_values: np.ndarray, mask: np.ndarray) -> np.ndarray:
        """Detect watermark presence in the given g-values.

        Args:
            g_values: Array of shape [batch_size, seq_len, watermarking_depth] containing
                the g-values computed from the text.
            mask: Binary array of shape [batch_size, seq_len] indicating which g-values
                should be used in detection. g-values with mask value 0 are discarded.

        Returns:
            Array of shape [batch_size] containing detection scores, where higher values
            indicate stronger evidence of watermarking.

        Raises:
            NotImplementedError: This is an abstract method that must be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement detect()")

class MeanDetector(SynthIDDetector):
    
    def detect(self, g_values, mask):
        """
        Args:
            g_values: shape [batch_size, seq_len, watermarking_depth]
            mask: shape [batch_size, seq_len]
        Returns:
            scores: shape [batch_size]
        """
        watermarking_depth = g_values.shape[-1]
        num_unmasked = np.sum(mask, axis=1)  # shape [batch_size]
        return np.sum(g_values * np.expand_dims(mask, 2), axis=(1, 2)) / (
                watermarking_depth * num_unmasked
        )
    

class WeightedMeanDetector(SynthIDDetector):
    
    def detect(
        self,
        g_values: np.ndarray,
        mask: np.ndarray,
        weights: np.ndarray = None,
    ) -> np.ndarray:
        """Computes the Weighted Mean score.

        Args:
            g_values: g-values of shape [batch_size, seq_len, watermarking_depth]
            mask: A binary array shape [batch_size, seq_len] indicating which g-values
                should be used. g-values with mask value 0 are discarded
            weights: array of non-negative floats, shape [watermarking_depth]. The
                weights to be applied to the g-values. If not supplied, defaults to
                linearly decreasing weights from 10 to 1

        Returns:
            Weighted Mean scores, of shape [batch_size]. This is the mean of the
            unmasked g-values, re-weighted using weights.
        """
        watermarking_depth = g_values.shape[-1]

        if weights is None:
            weights = np.linspace(start=10, stop=1, num=watermarking_depth)

        # Normalise weights so they sum to watermarking_depth
        weights *= watermarking_depth / np.sum(weights)

        # Apply weights to g-values
        g_values = g_values * np.expand_dims(weights, axis=(0, 1))

        num_unmasked = np.sum(mask, axis=1)  # shape [batch_size]
        return np.sum(g_values * np.expand_dims(mask, 2), axis=(1, 2)) / (
            watermarking_depth * num_unmasked
        )
    

class BayesianDetector(SynthIDDetector):
    def __init__(self, logits_processor):
        # Truncate outputs to this length for training.
        self.pos_truncation_length = 200
        self.neg_truncation_length = 200
        # Pad trucated outputs to this length for equal shape across all batches.
        self.max_padded_length = 1000
        self.trained_detector = None
        self.logits_processor = logits_processor
        self.device = torch.device(logits_processor.device)
        self.tokenizer = logits_processor.config.generation_tokenizer
        self.train_detector()
    
    def detect(self, g_values: np.ndarray, mask: np.ndarray) -> np.ndarray:
        """Detect watermark using Bayesian detector.
        
        Args:
            g_values: g-values of shape [batch_size, seq_len, watermarking_depth]
            mask: A binary array shape [batch_size, seq_len] indicating which g-values
                should be used. g-values with mask value 0 are discarded
                
        Returns:
            Detection scores of shape [batch_size]
        """
        if self.trained_detector is None:
            raise ValueError("Detector not trained. Call train_detector() first.")
        g_values = torch.tensor(g_values, dtype=torch.float32, device=self.device)
        mask = torch.tensor(mask, device=self.device)
        return self.trained_detector.detector_module(g_values, mask)
    

    def tokenize_and_batch_texts(self, texts, tokenizer, device, padded_length=2500, batch_size=16):
        tokenized_texts = []
        batched = []
        
        for text in tqdm.tqdm(texts):
            inputs = tokenizer(
                text,
                return_tensors='pt',
                add_special_tokens=False,
                padding=True,
            ).to(device)
            
            line = inputs['input_ids'].cpu().numpy()[0].tolist()
            if len(line) >= padded_length:
                line = line[:padded_length]
            else:
                line = line + [tokenizer.eos_token_id] * (padded_length - len(line))
            
            batched.append(torch.tensor(line, dtype=torch.long, device=device)[None, :])
            if len(batched) == batch_size:
                tokenized_texts.append(torch.cat(batched, dim=0))
                batched = []
        
            # Handle the last incomplete batch
        if batched:
            tokenized_texts.append(torch.cat(batched, dim=0))
        
        return tokenized_texts

    def get_data_for_training(self, dataset_name: str='c4', dataset_path: str='dataset/c4/processed_c4.json', num_samples: int=1000):
        unwatermarked_dataset = C4Dataset(dataset_path, max_samples=num_samples)
        watermarked_dataset = self.get_watermarked_data(unwatermarked_dataset)
        tokenized_natural_texts = self.tokenize_and_batch_texts(
            unwatermarked_dataset.natural_texts, 
            self.tokenizer, 
            self.device
        )

        # Process watermarked texts
        watermarked_texts = [item['watermarked_text'] for item in watermarked_dataset]
        tokenized_watermarked_texts = self.tokenize_and_batch_texts(
            watermarked_texts, 
            self.tokenizer, 
            self.device
        )
        return tokenized_natural_texts, tokenized_watermarked_texts


    def get_watermarked_data(self, dataset):
        """Generate and cache watermarked data from dataset prompts.
        
        Args:
            dataset: Dataset object containing prompts
            
        Returns:
            list: List of dictionaries containing prompts and their watermarked generations
        """
        output_dir = 'watermark/synthid/generated_texts'
        output_path = os.path.join(output_dir, 'watermarked_texts.json')
        
        # If cached data exists, load it
        if os.path.exists(output_path):
            with open(output_path, 'r', encoding='utf-8') as f:
                return json.load(f)
                
        # Create output directory if it doesn't exist
        os.makedirs(output_dir, exist_ok=True)
        
        # Generate watermarked texts
        generated_data = []
        for prompt in tqdm.tqdm(dataset.prompts):
            # Generate watermarked text
            watermarked_text = self.logits_processor.generate_watermarked_text(prompt)
            
            # Extract only the generated part (excluding prompt)
            generated_part = watermarked_text[len(prompt):].strip()
            
            # Store in list
            generated_data.append({
                'prompt': prompt,
                'watermarked_text': generated_part
            })

        # Save generated texts to JSON file
        with open(output_path, 'w', encoding='utf-8') as f:
            json.dump(generated_data, f, ensure_ascii=False, indent=2)
            
        return generated_data
        

    def train_detector(self):
        """Train the Bayesian detector.
        
        Args:
            wm_outputs: Watermarked outputs
            uwm_outputs: Unwatermarked outputs
            logits_processor: SynthID logits processor
            tokenizer: Tokenizer
            device: PyTorch device
        """
        uwm_outputs, wm_outputs = self.get_data_for_training()
        bayesian_detector, test_loss = (
            RawBayesianDetector.train_best_detector(
                tokenized_wm_outputs=wm_outputs,
                tokenized_uwm_outputs=uwm_outputs,
                logits_processor=self.logits_processor,
                tokenizer=self.tokenizer,
                torch_device=self.device,
                max_padded_length=self.max_padded_length,
                pos_truncation_length=self.pos_truncation_length,
                neg_truncation_length=self.neg_truncation_length,
                verbose=True,
                learning_rate=3e-3,
                n_epochs=50,
                l2_weights=np.zeros((1,)),
            )
        )
        self.trained_detector = bayesian_detector

def get_detector(detector_name: str, logits_processor):
    if detector_name == 'mean':
        return MeanDetector()
    elif detector_name == 'weighted_mean':
        return WeightedMeanDetector()
    elif detector_name == 'bayesian':
        return BayesianDetector(logits_processor)
    else:
        raise ValueError(f"Detector {detector_name} not found.")



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# ts.py
# Description: Implementation of TS algorithm
# ============================================

import torch
from functools import partial
from watermark.base import BaseWatermark, BaseConfig
from utils.utils import load_config_file
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
from visualize.data_for_visualization import DataForVisualization
from transformers import OPTForCausalLM, AutoTokenizer, LogitsProcessorList
from watermark.ts.TS_networks import DeltaNetwork, GammaNetwork

class TSConfig(BaseConfig):
    """Config class for TS algorithm, load config file and initialize parameters."""

    # def __init__(self, algorithm_config: str, transformers_config: TransformersConfig, *args, **kwargs) -> None:
    #     """
    #         Initialize the KGW configuration.

    #         Parameters:
    #             algorithm_config (str): Path to the algorithm configuration file.
    #             transformers_config (TransformersConfig): Configuration for the transformers model.
    #     """

    #     if algorithm_config is None:
    #         config_dict = load_config_file('config/TS.json')
    #     else:
    #         config_dict = load_config_file(algorithm_config)
    #     if config_dict['algorithm_name'] != 'TS':
    #         raise AlgorithmNameMismatchError('TS', config_dict['algorithm_name'])

    #     self.hash_key = config_dict['hash_key']
    #     self.seeding_scheme = config_dict['seeding_scheme']
    #     self.ckpt_path = config_dict['ckpt_path']
    #     self.gamma = config_dict['gamma']
    #     self.delta = config_dict['delta']
    #     self.prefix_length = config_dict['prefix_length']
    #     self.z_threshold = config_dict['z_threshold']
    #     self.tokenizer_opt = AutoTokenizer.from_pretrained("facebook/opt-1.3b", padding_side="left")

    #     self.generation_model = transformers_config.model
    #     self.generation_tokenizer = transformers_config.tokenizer
    #     self.vocab_size = transformers_config.vocab_size
    #     self.device = transformers_config.device
    #     self.gen_kwargs = transformers_config.gen_kwargs
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.gamma = self.config_dict['gamma']
        self.delta = self.config_dict['delta']
        self.hash_key = self.config_dict['hash_key']
        self.seeding_scheme = self.config_dict['seeding_scheme']
        self.ckpt_path = self.config_dict['ckpt_path']
        self.prefix_length = self.config_dict['prefix_length']
        self.z_threshold = self.config_dict['z_threshold']

        self.tokenizer_opt = AutoTokenizer.from_pretrained("facebook/opt-1.3b", padding_side="left")
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'TS'


class TSUtils:
    """Utility class for KGW algorithm, contains helper functions."""

    def __init__(self, config: TSConfig, *args, **kwargs) -> None:
        """
            Initialize the KGW utility class.

            Parameters:
                config (KGWConfig): Configuration for the KGW algorithm.
        """
        self.config = config
        self.device = self.config.device
        self.rng = torch.Generator(device=self.device)
        self.rng.manual_seed(self.config.hash_key)
        self.ckpt_path = self.config.ckpt_path
        self.seeding_scheme = self.config.seeding_scheme
        self.hash_key = self.config.hash_key
        self.vocab_size = self.config.vocab_size

        model = OPTForCausalLM.from_pretrained(
            "facebook/opt-1.3b",
            torch_dtype=torch.float,
            device_map='auto'
        )


        self.tokenizer_llama = self.config.generation_tokenizer
        self.embed_matrix = model.get_input_embeddings().weight.to(self.device)
        ckpt_path = self.config.ckpt_path

        if len(ckpt_path) > 0:
            print("checkpoint_load")
            checkpoint = torch.load(ckpt_path)
            layer_delta = sum(1 for key in checkpoint['delta_state_dict'] if "weight" in key)  # Counting only weight keys as layers
            layer_gamma = sum(1 for key in checkpoint['gamma_state_dict'] if "weight" in key)  # Counting only weight keys as layers

            self.gamma_network = GammaNetwork(input_dim=self.embed_matrix.shape[1], layers=layer_gamma).to(self.device)
            self.delta_network = DeltaNetwork(input_dim=self.embed_matrix.shape[1], layers=layer_delta).to(self.device)

            self.delta_network.load_state_dict(checkpoint['delta_state_dict'])
            self.gamma_network.load_state_dict(checkpoint['gamma_state_dict'])

            for name, param in self.delta_network.named_parameters():
                param.requires_grad = False
            for name, param in self.gamma_network.named_parameters():
                param.requires_grad = False
            self.delta_network.eval()
            self.gamma_network.eval()

        else:
            self.gamma = torch.tensor([self.config.gamma]).to(self.device)
            self.delta = torch.tensor([self.config.delta]).to(self.device)

        self.gamma_list = torch.empty(0, dtype=torch.float).to(self.device)
        self.delta_list = torch.empty(0, dtype=torch.float).to(self.device)

    def _seed_rng(self, input_ids: torch.LongTensor, seeding_scheme: str = None) -> None:
        # can optionally override the seeding scheme,
        # but uses the instance attr by default
        if seeding_scheme is None:
            seeding_scheme = self.seeding_scheme

        # using the last token and hash_key to generate random seed
        if seeding_scheme == "simple_1":
            assert input_ids.shape[-1] >= 1, f"seeding_scheme={seeding_scheme} requires at least a 1 token prefix sequence to seed rng"
            prev_token = input_ids[-1].item()
            self.rng.manual_seed(self.hash_key * prev_token)
        else:
            raise NotImplementedError(f"Unexpected seeding_scheme: {seeding_scheme}")
        return

    def _get_greenlist_ids(self, input_ids: torch.LongTensor,process) -> list[int]:
        # Always use ids given by OPT model
        # since our gamma/delta network is trained on the embedding matrix of OPT model
        # seed the rng using the previous tokens/prefix according to the seeding_scheme
        self._seed_rng(input_ids)

        # use last token to get gamma value and delta value
        if len(self.ckpt_path) > 0:
            gamma = self.gamma_network(self.embed_matrix[input_ids[-1].item()])
            delta = self.delta_network(self.embed_matrix[input_ids[-1].item()])
        else:
            delta = self.delta
            gamma = self.gamma

        if process == 'process':
        # get every token's gamma value and delta value
          self.gamma_list = torch.cat([self.gamma_list, gamma])
          self.delta_list = torch.cat([self.delta_list, delta])

        # generate greenlist, every token have different greenlist_id
        greenlist_size = int(self.vocab_size * gamma)
        vocab_permutation = torch.randperm(self.vocab_size, device=input_ids.device, generator=self.rng)

        greenlist_ids = vocab_permutation[:greenlist_size]

        gamma_len=len(self.gamma_list)

        return greenlist_ids, gamma, delta,gamma_len

    def _compute_z_score(self, observed_count, T):
        # count refers to number of green tokens, T is total number of tokens
        var = torch.sum(self.gamma_list * (1 - self.gamma_list))
        mean = torch.sum(self.gamma_list)
        z = (observed_count - mean)/torch.sqrt(var)
        return z

    def _score_sequence(
        self,
        input_ids: torch.Tensor,

    )-> tuple[float, list[int]]:

        num_tokens_scored = len(input_ids) - self.config.prefix_length
        if num_tokens_scored < 1:
            raise ValueError(
                (
                    f"Must have at least {1} token to score after "
                    f"the first min_prefix_len={self.config.prefix_length} tokens required by the seeding scheme."
                )
            )

        green_token_count = 0
        green_token_mask = [-1 for _ in range(self.config.prefix_length)]

        for idx in range(self.config.prefix_length, len(input_ids)):
            curr_token = input_ids[idx]
            if "opt" in self.config.generation_model.name_or_path.lower():
                greenlist_ids, gamma, delta, gamma_len = self._get_greenlist_ids(input_ids[:idx],"detect")

            else:
                llama_str = self.tokenizer_llama.decode(input_ids[max(idx-5, 0):idx], add_special_tokens=False)
                ids_opt = self.config.tokenizer_opt(llama_str, add_special_tokens=False)['input_ids']
                if len(ids_opt) == 0:
                    green_token_mask.append(False)
                    continue
                greenlist_ids, gamma, delta, gamma_len = self._get_greenlist_ids(torch.tensor(ids_opt).to(self.device),"detect")

            if curr_token in greenlist_ids:
                green_token_count += 1
                green_token_mask.append(True)
            else:
                green_token_mask.append(False)

        self.gamma_list = self.gamma_list[self.config.prefix_length:]

        z_score = self._compute_z_score(green_token_count, num_tokens_scored)

        return z_score, green_token_mask


class TSLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for TS algorithm, process logits to add watermark."""

    def __init__(self, config: TSConfig, utils: TSUtils, *args, **kwargs) -> None:

        self.config = config
        self.utils = utils

        self.tokenizer_llama = self.config.generation_tokenizer

        if "opt" not in self.config.generation_model.name_or_path.lower():
            self.vocab_size = len(self.tokenizer_llama)

    def _calc_greenlist_mask(self, logits: torch.FloatTensor, greenlist_token_ids) -> torch.BoolTensor:
        green_tokens_mask = torch.zeros_like(logits)
        green_tokens_mask[greenlist_token_ids] = 1
        final_mask = green_tokens_mask.bool()
        return final_mask

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:

        device = input_ids.device
        if self.utils.rng is None:
            self.utils.rng = torch.Generator(device=device)

        if "opt" not in self.config.generation_model.name_or_path.lower():
            llama_str = self.tokenizer_llama.batch_decode(input_ids[:,-5:], add_special_tokens=False)
            ids_opt = self.config.tokenizer_opt(llama_str, add_special_tokens=False)['input_ids']

        for b_idx in range(input_ids.shape[0]):

            if "opt" not in self.config.generation_model.name_or_path.lower():
                greenlist_ids, gamma, delta,gamma_len = self.utils._get_greenlist_ids(torch.tensor(ids_opt[b_idx]).to(device),'process')

            else:
                greenlist_ids, gamma, delta,gamma_len = self.utils._get_greenlist_ids(input_ids[b_idx],'process')

            # get greenlist token mask and add bias on the each logits base on greenlist mask
            green_tokens_mask = self._calc_greenlist_mask(logits=scores[b_idx], greenlist_token_ids=greenlist_ids)

            delta = delta.to(dtype=scores.dtype)
            scores[b_idx][green_tokens_mask] = scores[b_idx][green_tokens_mask] + delta

        return scores


class TS(BaseWatermark):
    """Top-level class for KGW algorithm."""

    def __init__(self, algorithm_config: str | TSConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the KGW algorithm.

            Parameters:
                algorithm_config (str | TSConfig): Path to the algorithm configuration file or TSConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = TSConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, TSConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a TSConfig instance")
        
        self.utils = TSUtils(self.config)
        self.logits_processor = TSLogitsProcessor(self.config, self.utils)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]),
            **self.config.gen_kwargs
        )

        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)

        prefix_length = encoded_prompt['input_ids'].shape[1]

        # Generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)

        # Decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text[:,prefix_length:], skip_special_tokens=True)[0]
        return watermarked_text

    def generate_unwatermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate unwatermarked text."""

        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)

        prefix_length = encoded_prompt['input_ids'].shape[1]

        # Generate unwatermarked text
        encoded_unwatermarked_text = self.config.generation_model.generate(**encoded_prompt, **self.config.gen_kwargs)
        # Decode
        unwatermarked_text = self.config.generation_tokenizer.batch_decode(encoded_unwatermarked_text[:,prefix_length:], skip_special_tokens=True)[0]
        return unwatermarked_text

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""

        # Encode the text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)

        # Compute z_score using a utility method
        z_score, _ = self.utils._score_sequence(encoded_text)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Convert z_score to a float
        z_score = z_score.item()
        is_watermarked = bool(is_watermarked)
                              
        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)

    def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
        """Get data for visualization."""

        # Encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)

        # Compute z-score and highlight values
        z_score, highlight_values = self.utils._score_sequence(encoded_text)

        # decode single tokens
        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)

        return DataForVisualization(decoded_tokens, highlight_values)